'''
PyPi Package: sbc_gpio
Homepage: https://www.learningtopi.com/python-modules-applications/sbc_gpio/
Git: https://github.com/LearningToPi/sbc_gpio

Description:
===========
This library is intended to provide an abstraction layer between the SBC and GPIO.  In particular
the Raspberry Pi group of devices uses a proprietary GPIO library (RPi.GPIO) that is not accessible
on other SBC platforms (for example Rock Pi, Banana Pi, Atomic Pi, etc).  This library is intended
to provide an abstract class that can allow for portable code between platforms.

The SBCPlatform class allows for identifying the platform based on platform specific criteria.  The
abstract GpioIn and GpioOut classes are used to provide access to GPIO libraries that are assigned
based on the identified platform.

Usage Example:
=============
from sbc_gpio import SBCPlatform, EVENT, DIR, PULL

# get the platform and associated GPIO library
platform = SBCPlatform()
print(platform)
print(platform.serial)

# Get a GPIO out and in pin
def callback_func(event:str, time:float, state:str)
    print(time, event, state)
gpio_out = platform.get_gpio_out(gpio_id='3B3', name='test_out', pull=PULL.NONE, initial_state=0)
gpio_in = platform.get_gpio_in(gpio_id='3B2', name='test_in', pull=PULL.DOWN, event=EVENT.BOTH, 
                               debounce_ms=100, callback=callback_func)
gpio_out.set_high()
gpio_out.set_low()


Folder Structure:
.\\DIR.py, .\\EVENT.py, .\\PULL.py:
    These files include constants that are used for consistent usage between GPIO libraries.

.\\platforms:
    This folder includes a file for each platform that can be identified by this library.  Each
    file should contain the following:
        MODEL_IDENTIFIER = [ <list of supported mechanisms for identifying the platform> ]
        PLATFORM_LOCAL = namedtuple('PLATFORM_LOCAL', (<tuple of platform specific values>))
        PLATFORM_SPECIFIC = PLATFORM_INFO(model=<short device model>, description=<long device model>,
                                          gpio_valid_values=[<list of valid integer gpio's>],
                                          local=PLATFORM_LOCAL(<instance of PLATFORM_LOCAL>))
        <optional>
        def convert_gpio(gpio_str:str) -> int:
            # Function to convert a passed string that is platform specific to a GPIO integer

        def convert_gpio_tuple(gpio_str:str) -> tuple:
            # Function to convert a passed string that is platform specific to a tuple that 
            #  contains the gpio chip and gpio integer.

.\\gpio_libs:
    This folder includes a "_generic_gpio.py" lib that represents a generic GPIO input or output.
    All GPIO library files in this folder will inherit these base clases.  The base classes allow
    the abstraction of the underlying GPIO library to make code more portable between SBC's.

    Each GPIO library file should override the GpioIn and GpioOut classes and provide GPIO class
    specific functionality.

.\\device_tests:
    This folder contains a series of test files based on the _test_base.py file.  These tests
    can be used to validate functionality on the different SBC platforms supported using the
    GPIO libraries available.

.\\tests:
    This folder contains the unittest files for running the SBC device tests.

.\\__main__.py:
    Running the module directly can be used to output information on the current device platform
    or execute tests on the platform.  See the __main__.py file for more details.

    
MIT License

Copyright (c) 2022 LearningToPi <contact@learningtopi.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
from importlib import import_module
import re
import sys
import subprocess
from collections import namedtuple
from logging_handler import create_logger, INFO
from . import DIR, EVENT, PULL
from .gpio_libs._generic_gpio import GpioIn, GpioOut
from .dynamic_dts import DynamicOverlay

PLATFORM_BASE_DIR = os.path.join(os.path.dirname(__file__), 'platforms')

PLATFORM_INFO = namedtuple('PLATFORM_INFO', ('model', 'description', 'gpio_valid_values',
                                             'dynamic_overlay_dir', 'dynamic_overlays', 'config_file',
                                             'update_extlinux_script', 'extlinux_conf', 'local'))

class SBCPlatform:
    ''' Class to handle platform specific commands on SBC '''
    def __init__(self, log_level=INFO):
        self._platform = None
        self._gpio_func = None
        self._logger = create_logger(console_level=log_level, name=self.info_str)
        self.serial = 'n/a'
        self.supported_platforms = []

        # Get a list of the platform files
        self._logger.debug(f"{self.info_str}: Listing platform files from: {PLATFORM_BASE_DIR}")
        platform_files = os.listdir(PLATFORM_BASE_DIR)
        mod = None
        for platform_file in platform_files:
            if platform_file.endswith('.py') and not platform_file.startswith('_'):
                try:
                    mod = import_module(f"sbc_gpio.platforms.{platform_file.split('.py')[0]}")
                    self.supported_platforms.append(mod.PLATFORM_SPECIFIC.model)
                    match_found = False
                    # run the platform identifier tests
                    for identifier in mod.MODEL_IDENTIFIER:
                        if identifier.get('type').lower() == 'true':
                            self._logger.debug(f"{self.info_str}: Platform matched in file {platform_file}")
                            # true identifier will always return true.  useful for testing
                            match_found = True
                            break

                        if identifier.get('type', 'file') == 'file' and os.path.exists(identifier['file']):
                            # file identifier will check the contents of the specified file for a provided regex
                            with open(identifier['file'], 'r', encoding='utf-8') as input_file:
                                file_contents = input_file.read()
                            if re.search(identifier['contents'], file_contents):
                                self._logger.debug(f"{self.info_str}: Platform matched in file {platform_file}")
                                match_found = True
                                break
                    # if we found a match, break out
                    if match_found:
                        break
                    # clear the loaded modules since we hit the end and didn't find a match
                    mod = None

                except Exception as e:
                    self._logger.error('Unable to import %s.  Error: %s', platform_file, e)

        # save the platform specific info and functions
        if mod is not None:
            self._platform = mod.PLATFORM_SPECIFIC
            if 'SERIAL_NUMBER' in dir(mod):
                with open(mod.SERIAL_NUMBER, 'r', encoding='utf-8') as input_file:
                    self.serial = input_file.read()
                    self.serial = ''.join([c for c in self.serial if str(c).isprintable()])
            if 'convert_gpio' in dir(mod):
                self.convert_gpio = mod.convert_gpio
            if 'convert_gpio_tuple' in dir(mod):
                self.convert_gpio_tuple = mod.convert_gpio_tuple
            self._platform_write_dynamic_overlay = mod.write_dynamic_overlay if 'write_dynamic_overlay' in dir(mod) else None
            self._compile_dtbo = mod.compile_dtbo if 'compile_dtbo' in dir(mod) else None
            self._GpioIn_Class = mod.GpioIn
            self._GpioOut_Class = mod.GpioOut
            self._logger.info(f"{self.info_str}: Platform identified as {self._platform.model} ({self.description})")
        else:
            raise ValueError(f"Unable to identify platform.  Supported devices are: {','.join(self.supported_platforms)}")

    def __str__(self):
        return self.model

    @property
    def model(self):
        ''' Return the model of the device as a string '''
        if isinstance(self._platform, PLATFORM_INFO):
            return self._platform.model
        return ''

    @property
    def description(self):
        ''' Return a description of the platform as a string '''
        if isinstance(self._platform, PLATFORM_INFO):
            return self._platform.description
        return ''

    @property
    def gpio_valid_values(self):
        ''' Return a list of valid GPIO values as a list '''
        if isinstance(self._platform, PLATFORM_INFO):
            return self._platform.gpio_valid_values
        return []

    @property
    def _local(self):
        ''' Return the platform local settings '''
        if isinstance(self._platform, PLATFORM_INFO):
            return self._platform.local
        return None

    def convert_gpio(self, *args) -> int: # pylint: disable=E0202
        ''' function placeholder - replaced with platform specific function, otherwise returns first parameter '''
        return args[0]

    def convert_gpio_tuple(self, *args) -> tuple: # pylint: disable=E0202
        ''' function placeholder - replaced with platform specific function, otherwise returns chip 0 and the first parameter as a tuple '''
        return 0, args[0]

    def gpio_is_valid(self, gpio_id) -> bool:
        ''' Checks if a passed value is a valid GPIO for the platform '''
        if isinstance(gpio_id, int) and gpio_id in self.gpio_valid_values:
            return True
        if self.convert_gpio(gpio_id) in self.gpio_valid_values:
            return True
        return False

    def get_gpio_out(self, gpio_id, name=None, pull=PULL.NONE, log_level=INFO, initial_state=0) -> GpioOut:
        ''' Get a gpio out pin.  Gpio_id can be a string (passed to convert), an int, or a tuple (chip, pin) '''
        gpio_tuple = tuple()
        if isinstance(gpio_id, int):
            gpio_tuple = (0, gpio_id)
        elif isinstance(gpio_id, str):
            gpio_tuple = self.convert_gpio_tuple(gpio_id)
        elif isinstance(gpio_id, tuple) and len(gpio_id) == 2:
            gpio_tuple = gpio_id
        else:
            raise ValueError('"gpio_id" must be an integer for a pin, or a tuple containing the gpio chip and gpio pin (chip, pin)')
        return self._GpioOut_Class(gpio_tuple[1], gpio_tuple[0], name=name, pull=pull, log_level=log_level, initial_state=initial_state)

    def get_gpio_in(self, gpio_id, name=None, pull=PULL.DOWN, event=EVENT.BOTH, debounce_ms=100, callback=None, log_level=INFO, start_polling=True) -> GpioIn:
        ''' Get a gpio in pin.  Gpio_id can be a string (passed to convert), an int, or a tuple (chip, pin) '''
        gpio_tuple = tuple()
        if isinstance(gpio_id, int):
            gpio_tuple = (0, gpio_id)
        elif isinstance(gpio_id, str):
            gpio_tuple = self.convert_gpio_tuple(gpio_id)
        elif isinstance(gpio_id, tuple) and len(gpio_id) == 2:
            gpio_tuple = gpio_id
        else:
            raise ValueError('"gpio_id" must be an integer for a pin, or a tuple containing the gpio chip and gpio pin (chip, pin)')
        return self._GpioIn_Class(gpio_tuple[1], gpio_tuple[0], name=name, pull=pull, event=event, debounce_ms=debounce_ms,
                                  callback=callback, log_level=log_level, start_polling=start_polling)

    @property
    def info_str(self):
        """ Returns the info string for the class (used in logging commands) """
        return f"{self.__class__.__name__}" + (f"({self._platform.model})" if self._platform is not None else '')

    def spi_buses(self) -> tuple:
        ''' Returns a tuple listing the spi bus numbers that are available (only applicable on Linux).  I.e. (0,1) or (0,) '''
        if 'linux' not in sys.platform.lower():
            raise NotImplementedError('spi_buses() only supported on Linux')
        dev_files = os.listdir('/dev')
        spi_buses = ()
        for dev_file in dev_files:
            if 'spidev' in dev_file:
                match = re.search("^spidev(?P<spi_bus>[0-9]).(?P<spi_cs>[0-9])$", dev_file)
                if match:
                    spi_buses += (int(match.group('spi_bus')),)
        return spi_buses


    def spi_bus_cs(self, spi_bus:int) -> tuple:
        ''' Returns a tuple listing the spi cs numbers that are available for the passed bus (only applicable on Linux).  I.e. (0,1) or (0,) '''
        if 'linux' not in sys.platform.lower():
            raise NotImplementedError('spi_buses() only supported on Linux')
        dev_files = os.listdir('/dev')
        spi_cs = ()
        for dev_file in dev_files:
            if 'spidev' in dev_file:
                match = re.search("^spidev(?P<spi_bus>[0-9]).(?P<spi_cs>[0-9])$", dev_file)
                if match and int(match.group('spi_bus')) == spi_bus:
                    spi_cs += (int(match.group('spi_cs')),)
        return spi_cs


    def i2c_buses(self) -> tuple:
        ''' Returns a tuple listing the i2c bus numbers that are available (only applicable on Linux).  I.e (0,1,5,7) '''
        if 'linux' not in sys.platform.lower():
            raise NotImplementedError('spi_buses() only supported on Linux')
        dev_files = os.listdir('/dev')
        i2c_buses = ()
        for dev_file in dev_files:
            if 'i2c-' in dev_file:
                match = re.search("^i2c-(?P<i2c_bus>[0-9])$", dev_file)
                if match:
                    i2c_buses += (int(match.group('i2c_bus')),)
        return i2c_buses

    def write_dynamic_overlays(self):
        ''' Read any dynamic overlays from the config file and write dts and dtbo files'''
        # check that dynamic DTS is supported
        if self._platform is None or not isinstance(self._platform.dynamic_overlay_dir, str) or not isinstance(self._platform.dynamic_overlays, dict) or len(self._platform.dynamic_overlays) == 0:
            self._logger.warning(f"{self.info_str}: Platform does not support dynamic overlay configuration.")
            self._logger.info(f"{self.info_str}: Dynamic overlay configuration aborted.")
            return

        # create dynamic_overlay folder
        try:
            os.makedirs(self._platform.dynamic_overlay_dir, mode=0o755, exist_ok=True)

            # Delete any existing dynamic overlay files (will recreate ones that still exist in the config)
            dynamic_overlay_files = os.listdir(self._platform.dynamic_overlay_dir)
            dynamic_overlay_files = [f for f in dynamic_overlay_files if os.path.isfile(self._platform.dynamic_overlay_dir+'/'+f)]
            self._logger.debug(f"{self.info_str}: Files currently in {self._platform.dynamic_overlay_dir}: {dynamic_overlay_files}")
            for file in dynamic_overlay_files:
                self._logger.debug(f"{self.info_str}: Deleting file {file}")
                os.remove(os.path.join(self._platform.dynamic_overlay_dir, file))
        except PermissionError:
            self._logger.error(f"{self.info_str}: Insufficient permisions to write to {self._platform.dynamic_overlay_dir}. Check permissions or run with sudo!")
            self._logger.info(f"{self.info_str}: Dynamic overlay configuration aborted.")
            return

        # Read all dynamic overlays from the config file
        dynamic_overlays = {}
        try:
            with open(self._platform.config_file, 'r', encoding='utf-8') as input_file:
                config_lines = input_file.readlines()
            for config_line in config_lines:
                dynamic_dts_re = re.search(r'^dynamic_overlay=(?P<driver>[^\s,]*),*(?P<options>[^\s]*)', config_line)
                if dynamic_dts_re:
                    overlay_name = dynamic_dts_re.group('driver')
                    overlay_params = dynamic_dts_re.group('options')
                    self._logger.debug(f"{self.info_str}: {self._platform.config_file}: Found dynamic overlay {overlay_name}, passed params: {overlay_params}")

                    # parse the overlay parameters, comma after overlay name and between options.  = between option name and value.  If no value, assume TRUE if present
                    _tmp_params = overlay_params.split(',')
                    overlay_params = {}
                    for _param in _tmp_params:
                        if _param != '':
                            _param_split = _param.split('=')
                            overlay_params[_param_split[0]] = _param_split[1] if len(_param_split) > 1 else True

                    # update overlay with config
                    if overlay_name not in self._platform.dynamic_overlays:
                        self._logger.warning(f"{self.info_str}: Overlay {overlay_name} not in list of supported dynamic overlays: {self._platform.dynamic_overlays.keys()}")
                        continue
                    if overlay_name not in dynamic_overlays:
                        # we haven't configured the overlay yet, so pull it in and update the parameters
                        dynamic_overlays[overlay_name] = DynamicOverlay(name=overlay_name, params=self._platform.dynamic_overlays[overlay_name]['params'])
                    try:
                        dynamic_overlays[overlay_name].set_params(overlay_params)
                    except ValueError as e:
                        self._logger.error(f"{self.info_str}: Error configuring overlay {overlay_name}: {e}")
                        continue

            # verify that all configured dynamic overlays have all required configurations
            for key, item in dynamic_overlays.items():
                if not item.ok:
                    self._logger.error(f"{self.info_str}: Overlay {key} is missing required parameters: {item.missing_params}")
                    for missing_param in item.missing_params:
                        self._logger.info(f"    Param {missing_param} help: {item.get_param(missing_param).param_help}")
                    continue

                # write the DTS file
                if self._platform_write_dynamic_overlay is not None:
                    self._logger.info(f"{self.info_str}: Writing dynamic overlay " + os.path.join(self._platform.dynamic_overlay_dir, f"{item.name}-{item.get_param('gpio').set_value}.dts"))
                    self._platform_write_dynamic_overlay(item)
                else:
                    self._logger.error(f"{self.info_str}: Unable to write overlay {key}.  Missing required platform function.")

                # compile DTBO
                if self._platform_write_dynamic_overlay is not None:
                    self._logger.info(f"{self.info_str}: Compiling dynamic overlay " + os.path.join(self._platform.dynamic_overlay_dir, f"{item.name}-{item.get_param('gpio').set_value}.dtbo"))
                    dts_file = os.path.join(self._platform.dynamic_overlay_dir, f"{item.name}-{item.get_param('gpio').set_value}.dts")
                    dtbo_file = os.path.join(self._platform.dynamic_overlay_dir, f"{item.name}-{item.get_param('gpio').set_value}.dtbo")
                    subprocess.run(f"dtc -O dtb -o {dtbo_file} {dts_file}", check=True, shell=True)
                    # verify file exists
                    if not os.path.exists(dtbo_file):
                        raise TimeoutError(f'Error creating DTBO {dtbo_file}')
                else:
                    self._logger.error(f"{self.info_str}: Unable to write overlay {key}.  Missing required platform function.")

            # Run update extlinux script if required
            if self._platform.update_extlinux_script is not None:
                self._logger.info(f"{self.info_str}: Running extlinux update script {self._platform.update_extlinux_script}")
                subprocess.run(self._platform.update_extlinux_script, check=True)

            if self._platform.extlinux_conf is not None:
                for key, item in dynamic_overlays.items():
                    # add dynamic overlay to extlinux config
                    self._logger.info(f"{self.info_str}: Updating extlinux config file {self._platform.extlinux_conf}")
                    dtbo_file = os.path.join(self._platform.dynamic_overlay_dir, f"{item.name}-{item.get_param('gpio').set_value}.dtbo").replace('/boot', '')
                    with open(self._platform.extlinux_conf, 'r', encoding='utf-8') as input_file:
                        extlinux_conf = input_file.readlines()
                    # loop through and find each label point
                    label_lines = [line for line in extlinux_conf if line.startswith('label')]
                    fdtoverlays_lines = [line for line in extlinux_conf if re.search(r'\s*fdtoverlays\s*.*', line)]
                    if len(fdtoverlays_lines) == 0:
                        # there are no fdtoverlay settings, so we need to add it in after the label line
                        for label_line in label_lines:
                            self._logger.debug(f"{self.info_str}: Adding new fdtoverlays line after label in extlinux.conf")
                            extlinux_conf.insert(extlinux_conf.index(label_line) + 1, f"    fdtoverlays {dtbo_file}\n")
                    else:
                        for fdtoverlays_line in fdtoverlays_lines:
                            self._logger.debug(f"{self.info_str}: adding dynamic overlay to existng fdtoverlays in extlinux.conf")
                            extlinux_conf[extlinux_conf.index(fdtoverlays_line)] = extlinux_conf[extlinux_conf.index(fdtoverlays_line)].rstrip() + f" {dtbo_file}\n"

                    self._logger.info(f"{self.info_str}: Writing updated extlinux configuration file {self._platform.extlinux_conf}")
                    with open(self._platform.extlinux_conf, 'w', encoding='utf-8') as output_file:
                        output_file.writelines(extlinux_conf)

        except Exception as e:
            self._logger.error(f"{self.info_str}: Error creating dynamic overlays: {e}")



from ._test_base import DevTest_Base, DevTest_Results, DEBUG, INFO, ERROR, WARNING, CRITICAL
import os
import subprocess
import socket
import lirc
from collections import namedtuple
from time import time, sleep
import re

TEST_NAME = 'IR GPIO TX/RX'
TEST_DESCRIPTION = 'Uses lirc with an infrared LED and receiver to test sending and receiving IR signals.'
PASS_THRESHOLD = 0.75

LIRC_DEVS = namedtuple("LIRC_DEVS", ("tx", "rx"))
LIRC_REMOTE = namedtuple("LIRC_REMOTE", ('name', 'keys'))

LIRC_DEVICE_PATH = '/tmp/'

LIRC_REMOTE_FILE = os.path.join(os.path.dirname(__file__), 'lirc', 'aa59-00741a.lircd.conf')
LIRC_EXCLUDE_REMOTES = ['devinput-32', 'devinput-64']


def get_pid_list(filter_list:list) -> list:
    ''' Get a list of running pids and return the pid matching the list of regex filters '''
    ps_list = subprocess.run('ps aux'.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return_pids = [None] * len(filter_list)
    for process_line in ps_list.stdout.decode('utf-8').split('\n'):
        for filter in filter_list:
            if re.search(filter, process_line):
                return_pids[filter_list.index(filter)] = process_line.split()[1] # type: ignore
    return return_pids

class DevTest_IR(DevTest_Base):
    ''' Class for an IR TX/RX test using lirc and GPIO based IR transmitter and receiver '''
    def __init__(self, log_level=INFO, lircd_path=LIRC_DEVICE_PATH):
        super().__init__(log_level)
        self._ir_ok = None
        self._ir_tx_obj = None
        self._ir_rx_obj = None
        self.ir_tx_dev = None
        self.ir_rx_dev = None
        self.ir_tx_driver = None
        self.ir_rx_driver = None
        self._ir_tx_process = None
        self._ir_rx_process = None
        self.lircd_path = lircd_path

    def close(self):
        # kill lircd processes if started by devtester
        if self._ir_tx_process is not None:
            lircd_procs = get_pid_list([f"lircd.*{self.ir_tx_dev}"])
            for process in lircd_procs:
                if process is not None:
                    self._logger.info(f"{self.info_str}: lircd TX process pid: {self._ir_tx_process.pid} stopping...")
                    subprocess.run(f'kill {process}'.split())
        if self._ir_rx_process is not None:
            lircd_procs = get_pid_list([f"lircd.*{self.ir_rx_dev}"])
            for process in lircd_procs:
                if process is not None:
                    self._logger.info(f"{self.info_str}: lircd RX process pid: {self._ir_rx_process.pid} stopping...")
                    subprocess.run(f'kill {process}'.split())
        sleep(1)
        if self._lircd_rx_proc() or self._lircd_tx_proc():
            raise RuntimeError(f'{self.info_str}: Error terminating lircd processes!')

    @property
    def _ir_tx(self):
        ''' Return the TX LIRC object '''
        if isinstance(self._ir_tx_obj, lirc.Client):
            return self._ir_tx_obj
        if self.is_test_ok:
            self._ir_tx_obj = lirc.Client(lirc.LircdConnection(address=self.lircd_path + self.ir_tx_dev)) # type: ignore
            return self._ir_tx_obj
        raise LookupError(f'{self.info_str}: Unable to open TX IR device')

    @property
    def _ir_rx(self):
        ''' Return the RX socket '''
        if isinstance(self._ir_rx_obj, socket.socket):
            return self._ir_rx_obj
        if self.is_test_ok:
            self._ir_rx_obj = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._ir_rx_obj.connect(self.lircd_path + self.ir_rx_dev) # type: ignore
            self._ir_rx_obj.setblocking(False)
            return self._ir_rx_obj
        raise LookupError(f'{self.info_str}: Unable to open RX IR device')
    
    @property
    def _remote_codes(self) -> tuple:
        ''' Return a tuple containing the remote and keys for all remotes available to LIRCD and associated tuple containing all codes '''
        raw_remotes = self._ir_tx.list_remotes()
        if isinstance(raw_remotes, list):
            remotes = tuple([LIRC_REMOTE(x, tuple([y.split()[1] for y in self._ir_tx.list_remote_keys(x)])) for x in raw_remotes if x not in LIRC_EXCLUDE_REMOTES])
        elif isinstance(raw_remotes, str):
            remotes = tuple([LIRC_REMOTE(raw_remotes, tuple([y.split()[1] for y in self._ir_tx.list_remote_keys(raw_remotes)]))])
        else:
            raise ValueError(f'Unable to query remotes from {self.ir_tx_dev}.  Received a {type(raw_remotes)}, expected a list or string.')
        if len(remotes) == 0:
            raise ValueError(f"{self.info_str}: Unable to identify any configured remote controls")
        return remotes

    @property
    def is_test_ok(self):
        ''' Checks if there are 2 IR devices, one using send and one receive '''
        if self._ir_ok is not None:
            return self._ir_ok
        try:
            lirc_sys = os.listdir('/sys/class/lirc')
            if len(lirc_sys) < 2:
                self._logger.warning(f"{self.info_str}: Insufficient lirc devices.  Skipping all IR tests.")
                self._ir_ok = False
            for lirc_dev in lirc_sys:
                driver_path = os.path.realpath(f'/sys/class/lirc/{lirc_dev}/device/device')
                driver_name = driver_path.rsplit('/', 1)[1]
                device = {'os_driver': driver_name}
                if 'tx' in driver_name or 'transmit' in driver_name:
                    device['dir'] = 'out'
                    self.ir_tx_dev = lirc_dev
                    self.ir_tx_driver = driver_name
                elif 'rx' in driver_name or 'recv' in driver_name or 'receiver' in driver_name:
                    device['dir'] = 'in'
                    self.ir_rx_dev = lirc_dev
                    self.ir_rx_driver = driver_name
                else:
                    raise AttributeError(f'Unable to determine direction of {lirc_dev}')
            
            # make sure there is at least 1 send and 1 recieve device
            if self.ir_tx_dev is not None and self.ir_rx_dev is not None:
                self._ir_ok = True
                return True
            return False

        except Exception as e:
            self._logger.warning(f"{self.info_str}: Unable to list lirc devices.  Skipping all IR tests: {e}")
            self._ir_ok = False

    def _lircd_tx_proc(self) -> bool:
        ''' Check if lircd tx process started by the devtest python is up '''
        if not self.is_test_ok:
            return False
        if self._ir_tx_process and self._ir_tx_process.poll() is None:
            # process started by devtest
            return True
        if get_pid_list([f"lircd.*{self.ir_tx_dev}"])[0] != None:
            # other lircd process started outside of devtest for the device specified
            return True
        return False
        
    def _lircd_rx_proc(self) -> bool:
        ''' Check if lircd rx process started by the devtest python is up '''
        if not self.is_test_ok:
            return False
        if self._ir_rx_process and self._ir_rx_process.poll() is None:
            return True
        if get_pid_list([f"lircd.*{self.ir_rx_dev}"])[0] != None:
            # other lircd process started outside of devtest for the device specified
            return True
        return False
    
    def lirc_start(self):
        ''' Starts the lirc daemon for the TX and RX '''
        if not self.is_test_ok:
            self._logger.warning(f"{self.info_str}: Unable to start IR devices.  All IR tests skipped.")
            return False
        
        # check if lirc processes are already running
        try:
            if not self._lircd_tx_proc():
                # start the lircd TX process
                self._logger.info(f"{self.info_str}: lircd TX process starting...")
                self._ir_tx_process = subprocess.Popen(f"/sbin/lircd --driver=default --device=/dev/{self.ir_tx_dev} --output={self.lircd_path + self.ir_tx_dev} --pidfile={self.lircd_path + self.ir_tx_dev}.pid --nodaemon {LIRC_REMOTE_FILE}".split(),  # type: ignore
                                                               stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                # wait 1 second and make sure the process didn't terminate
                sleep(1)
                if self._ir_tx_process.poll() != None:
                    stdout, stderr = self._ir_tx_process.communicate()
                    self._logger.error(f"{self.info_str}: Error starting LIRC TX process. code: {self._ir_tx_process.poll()}, stdout: {stdout}, stderr: {stderr}")
                    self._ir_tx_process = None
                else:
                    self._logger.info(f"{self.info_str}: lircd TX process pid: {self._ir_tx_process.pid}")
                
            if not self._lircd_rx_proc():
                # start the lircd RX process
                self._logger.info(f"{self.info_str}: lircd RX process starting...")
                self._ir_rx_process = subprocess.Popen(f"/sbin/lircd --driver=default --device=/dev/{self.ir_rx_dev} --output={self.lircd_path + self.ir_rx_dev} --pidfile={self.lircd_path + self.ir_rx_dev}.pid --nodaemon {LIRC_REMOTE_FILE}".split(),  # type: ignore
                                                               stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                # wait 1 second and make sure the process didn't terminate
                sleep(1)
                if self._ir_rx_process.poll() != None:
                    stdout, stderr = self._ir_rx_process.communicate()
                    self._logger.error(f"{self.info_str}: Error starting LIRC TX process. code: {self._ir_rx_process.poll()}, stdout: {stdout}, stderr: {stderr}")
                    self._ir_rx_process = None
                else:
                    self._logger.info(f"{self.info_str}: lircd RX process pid: {self._ir_rx_process.pid}")

            if not self._lircd_tx_proc() or not self._lircd_rx_proc():
                raise ConnectionError(f"{self.info_str}: Unable to start lircd IR processes")
        
        except Exception as e:
            raise ConnectionError(f"{self.info_str}: Unable to start lircd IR processes")

    def lirc_stop(self, force=False):
        ''' Stops the lirc daemon for the TX and RX. '''
        self.close()

    def _transmit_blocking(self, remote, key, tx_ms, wait_delay_ms=100) -> str:
        ''' Send the IR command and validate response.  Return received string '''
        self._logger.debug(f"{self.info_str}: IR Test: Sending {remote} {key}")
        self.lirc_rx_read_code() # Read from the RX to clear any stored codes
        self._ir_tx.send_start(remote, key)
        sleep(tx_ms / 1000)
        self._ir_tx.send_stop(remote, key)
        sleep(wait_delay_ms / 1000)
        recv = self.lirc_rx_read_code()
        self._logger.debug(f"{self.info_str}: IR Test: Received {recv}")
        return recv

    def _start_thread_iterations(self, iterations, tx_ms=1000, delay_ms=250, run_secs=None):
        ''' Run the IR test threaded or blocking '''
        with self._testing_lock:
            self.lirc_start()
            remote_list = self._remote_codes
            iter_run, iter_success = 0, 0
            self.lirc_rx_read_code() # read from the RX stream to clear any pending data
            start_time = time()
            try:
                while iter_run < iterations and not self._stop_tests:
                    for remote in remote_list:
                        for key in remote.keys:
                            iter_run += 1
                            recv = self._transmit_blocking(remote.name, key, tx_ms)
                            if recv == key:
                                iter_success += 1
                            else:
                                self._logger.warning(f"{self.info_str}: IR Test: Failed reading {remote.name} {key}.  Recived: {recv}")
                            if self._stop_tests or iter_run >= iterations:
                                raise StopIteration
                            sleep(delay_ms / 1000)
            except StopIteration:
                pass
            except Exception as e:
                self._logger.error(f"{self.info_str}: IR Test: Error during test loop: {e}")
            end_time = time()
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success, 
                                                parameters={'iterations': iterations, 'tx_ms': tx_ms, 'delay_ms':delay_ms,
                                                            'ir_tx_driver': self.ir_tx_driver, 'ir_rx_driver': self.ir_rx_driver},
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION)

    def _start_thread_time(self, run_secs=10, tx_ms=1000, delay_ms=250, iterations=None):
        ''' Run the IR test threaded or blocking '''
        with self._testing_lock:
            self.lirc_start()
            remote_list = self._remote_codes
            iter_run, iter_success = 0, 0
            self.lirc_rx_read_code() # read from the RX stream to clear any pending data
            start_time = time()
            stop_time = time() + run_secs
            try:
                while time() < stop_time and not self._stop_tests:
                    for remote in remote_list:
                        for key in remote.keys:
                            iter_run += 1
                            recv = self._transmit_blocking(remote.name, key, tx_ms)
                            if recv == key:
                                iter_success += 1
                            else:
                                self._logger.warning(f"{self.info_str}: IR Test: Failed reading {remote.name} {key}.  Recived: {recv}")
                            if self._stop_tests or time() >= stop_time:
                                raise StopIteration
                            sleep(delay_ms / 1000)
            except StopIteration:
                pass
            except Exception as e:
                self._logger.error(f"{self.info_str}: IR Test: Error during test loog: {e}")
            end_time = time()
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success, 
                                                parameters={'run_secs': run_secs, 'tx_ms': tx_ms, 'delay_ms':delay_ms,
                                                            'ir_tx_driver': self.ir_tx_driver, 'ir_rx_driver': self.ir_rx_driver},
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION)

    def lirc_rx_read_code(self) -> str:
        ''' Read any pending data from the socket and return the last code received '''
        try:
            data = self._ir_rx.recv(8192)
            return data.decode('utf-8').strip().split('\n')[-1].split()[-2]
        except BlockingIOError:
            return 'NO-DATA'


''' 
General GPIO configuration constants and set function using general linux kernel configuration:
https://www.kernel.org/doc/Documentation/devicetree/bindings/gpio/gpio.txt

Override on a per device basis as needed.
'''

ACTIVE_HIGH =     0b0 # bit 0
ACTIVE_LOW =      0b1 # bit 0
PUSH_PULL =       0b0 # bit 1
SINGLE_ENDED =    0b1 # bit 1
OPEN_SOURCE =     0b0 # bit 2
OPEN_DRAIN =      0b1 # bit 2
SLEEP_MAINTAIN =  0b0 # bit 3
SLEEP_STATE_LOST= 0b1 # bit 3
PULL_UP =         0b01 # bits 4 and 5
PULL_DOWN =       0b10 # bits 4 and 5
PULL_NONE =       0b00 # bits 4 and 5


FLAGS = {
    'ACTIVE_HIGH':       {'value': 0b0, 'offset': 0}, 
    'ACTIVE_LOW':        {'value': 0b1, 'offset': 0}, 
    'PUSH_PULL':         {'value': 0b0, 'offset': 1}, 
    'SINGLE_ENDED':      {'value': 0b1, 'offset': 1}, 
    'OPEN_SOURCE':       {'value': 0b0, 'offset': 2},
     'OPEN_DRAIN':       {'value': 0b1, 'offset': 2}, 
     'SLEEP_MAINTAIN':   {'value': 0b0, 'offset': 3}, 
     'SLEEP_STATE_LOST': {'value': 0b1, 'offset': 3}, 
     'PULL_UP':          {'value': 0b01, 'offset': 4}, 
     'PULL_DOWN':        {'value': 0b10, 'offset': 4}, 
     'PULL_NONE':        {'value': 0b00, 'offset': 4}
     }

def set_gpio_flags(*args) -> int:
    ''' Return an integer based on the flags sent.'''
    ret_value = 0
    for arg in args:
        if arg.upper() not in FLAGS.keys():
            raise ValueError(f'Value {arg} not in list of supported flags: {FLAGS}')
        ret_value = ret_value | (FLAGS[arg.upper()].get('value', 0b0) << FLAGS[arg.upper()].get('offset', 0))
    return ret_value


'''
Base GPIO classes used to represent a in and out GPIO and provide common functionality
across all gpio libraries.
'''
from logging_handler import create_logger
from threading import Thread, Lock
from sbc_gpio import EVENT

class Gpio:
    ''' Base GPIO functions that can be used for all input or output GPIO's '''
    def __init__(self, name, log_level):
        self.name = name if name is not None else f"Generic GPIO"
        self._logger = create_logger(console_level=log_level, name=self.__class__.__name__)

    def __del__(self):
        self.close()

    def close(self):
        pass

    @property
    def info_str(self):
        ''' Returns the info string for the class (used in logging commands) '''
        return f"{self.__class__.__name__} ({self.name})"

class GpioOut(Gpio):
    ''' Base GPIO class to represent a GPIO configured for output '''
    def __init__(self, name, log_level, pull):
        super().__init__(name, log_level)
        self.pull = pull

    def state(self):
        pass

    def set_high(self):
        pass

    set_1 = set_high
    set_on = set_high

    def set_low(self):
        pass

    set_0 = set_low
    set_off = set_low


class GpioIn(Gpio):
    ''' Base GPIO class to represent a GPIO configured for input '''
    def __init__(self, name, log_level, event, callback, debounce_ms, pull):
        super().__init__(name, log_level)
        self._edge_thread = None
        self._stop_thread = False
        self.event = event
        self.callback = callback
        self.debounce_ms = debounce_ms
        self.pull = pull

    def stop(self):
        ''' Stop background event polling '''
        if isinstance(self._edge_thread, Thread) and self._edge_thread.is_alive():
            self._logger.info(f"{self.info_str}: Stopping event thread...")
            self._stop_thread = True
            self._edge_thread.join()

    def start(self):
        ''' Start background event polling '''
        self.stop()
        if self.event == EVENT.RISING or self.event == EVENT.FALLING or self.event == EVENT.BOTH:
            self._logger.info(f"{self.info_str}: Starting event thread...")
            self._edge_thread = Thread(target=self._event_thread, name=f'gpiod-in-thread', daemon=True)
            self._edge_thread.start()

    @property
    def state(self):
        ''' Return current state of the input pin '''
        pass

    @property
    def event_thread_running(self):
        ''' Return True/False if the background event thread is running '''
        if isinstance(self._edge_thread, Thread) and self._edge_thread.is_alive():
            return True
        return False
    
    def _event_thread(self):
        ''' Backgroun thread to watch for rising or falling edges '''
        pass


'''
GPIO library:  gpiod
Supported platforms: Most modern SBC devices that support the libgpiod kernel driver
'''
import gpiod
from threading import Thread
from time import time
from datetime import timedelta
from sbc_gpio import PULL, EVENT
from ._generic_gpio import GpioIn as Generic_GpioIn, GpioOut as Generic_GpioOut
from logging_handler import INFO


class GpioOut(Generic_GpioOut):
    ''' Class to represent an abstracted GPIO pin using the gpiod '''
    def __init__(self, gpio_pin, gpio_chip, name=None, pull=PULL.NONE, log_level=INFO, initial_state=0):
        super().__init__(name=name, log_level=log_level, pull=pull)
        self.name = name if name is not None else f"chip:{gpio_chip},pin:{gpio_pin}"
        self.gpio_pin, self.gpio_chip = gpio_pin, gpio_chip
        # initialize the chip and pin
        chip = gpiod.chip(int(gpio_chip), gpiod.chip.OPEN_BY_NUMBER)
        self._pin = chip.get_line(int(self.gpio_pin))
        pin_config = gpiod.line_request()
        pin_config.consumer = name if name is not None else f'{self.info_str}-OUT'
        pin_config.request_type = gpiod.line_request.DIRECTION_OUTPUT
        if pull == PULL.UP:
            pin_config.flags = gpiod.line_request.FLAG_BIAS_PULL_UP
        elif pull == PULL.DOWN:
            pin_config.flags = gpiod.line_request.FLAG_BIAS_PULL_DOWN
        else:
            pin_config.flags = gpiod.line_request.FLAG_BIAS_DISABLE
        self._logger.info(f"{self.info_str}: Requesting GPIO...")
        self._pin.request(pin_config)

        if initial_state == 0:
            self.set_0()
        else:
            self.set_1()

    def close(self):
        self._logger.info(f"{self.info_str}: Releasing GPIO...")
        self._pin.release()

    @property
    def state(self):
        ''' Return current CS state '''
        return self._pin.get_value()

    def set_high(self):
        ''' Set the pin to on/high '''
        self._pin.set_value(1)
    
    def set_low(self):
        ''' Set the pin to off/low '''
        self._pin.set_value(0)

class GpioIn(Generic_GpioIn):
    ''' Class to represent an abstracted GPIO pin using the gpiod '''
    def __init__(self, gpio_pin, gpio_chip, name=None, pull=PULL.DOWN, event=EVENT.BOTH, debounce_ms=100, callback=None, log_level=INFO, start_polling=True):
        super().__init__(name=name, log_level=log_level, event=event, callback=callback, debounce_ms=debounce_ms, pull=pull)
        self.name = name if name is not None else f"chip:{gpio_chip},pin:{gpio_pin}"
        self.gpio_pin, self.gpio_chip = gpio_pin, gpio_chip
        # initialize the chip and pin
        chip = gpiod.chip(str(gpio_chip), gpiod.chip.OPEN_BY_NUMBER)
        self._pin = chip.get_line(int(self.gpio_pin))
        pin_config = gpiod.line_request()
        pin_config.consumer = name if name is not None else f'{self.info_str}-IN'
        pin_config.request_type = gpiod.line_request.DIRECTION_INPUT
        if pull == PULL.UP:
            pin_config.flags = gpiod.line_request.FLAG_BIAS_PULL_UP
        elif pull == PULL.DOWN:
            pin_config.flags = gpiod.line_request.FLAG_BIAS_PULL_DOWN
        else:
            pin_config.flags = gpiod.line_request.FLAG_BIAS_DISABLE
        # set edge request - filter in loop
        pin_config.request_type = gpiod.line_request.EVENT_BOTH_EDGES
        self._logger.info(f"{self.info_str}: Requesting GPIO...")
        self._pin.request(pin_config)
        self._stop_thread = False
        self._edge_thread = None
        if start_polling:
            self.start()

    def close(self):
        self.stop()
        self._logger.info(f"{self.info_str}: Releasing GPIO...")
        self._pin.release()

    @property
    def state(self):
        ''' Return current CS state '''
        return self._pin.get_value()
    
    def _event_thread(self): # type: ignore
        ''' Background thread to watch for rising or falling edge '''
        triggered = False

        def call_event(event, triggered):
            if self.callback is not None:
                self._logger.debug(f"{self.info_str}: {EVENT.RISING.upper() if event.event_type == gpiod.line_event.RISING_EDGE else EVENT.FALLING.upper()} state: {triggered}")
                Thread(target=self.callback, kwargs={'event': EVENT.RISING if event.event_type == gpiod.line_event.RISING_EDGE else EVENT.FALLING,
                      'time': time(), 'state': triggered}).start()
            else:
                self._logger.info(f"{self.info_str}: {EVENT.RISING.upper() if event.event_type == gpiod.line_event.RISING_EDGE else EVENT.FALLING.upper()} state: {triggered}")

        while True and not self._stop_thread:
            try:
                event_triggered = self._pin.event_wait(timedelta(seconds=1))
                if event_triggered:
                    event = self._pin.event_read()
                    # check for another event within the debounce interval
                    bounce_event_triggered = self._pin.event_wait(timedelta(milliseconds=self.debounce_ms))
                    if not bounce_event_triggered:
                        # if monitoring for event rising or event falling ONLY, just send the event
                        if self.event != EVENT.BOTH:
                            if (self.event == EVENT.RISING and event.event_type == gpiod.line_event.RISING_EDGE) or (self.event == EVENT.FALLING and event.event_type == gpiod.line_event.FALLING_EDGE):
                                triggered = True
                                call_event(event, triggered)
                        # otherwise need to track when we are triggered so we don't alert to a release if there was no press!
                        else:
                            if (event.event_type == gpiod.line_event.RISING_EDGE and self.pull == PULL.DOWN) or (event.event_type == gpiod.line_event.FALLING_EDGE and self.pull != PULL.DOWN):
                                triggered = True
                                call_event(event, triggered)
                            elif triggered:
                                triggered = False
                                call_event(event, triggered)
                            
                    event = None
            except Exception as e:
                self._logger.error(f"{self.info_str}: Error in event thread: {e}. Restarting...")
                triggered = False
        # reset the stop thread variable
        self._stop_thread = False



from ._test_base import DevTest_Base, DevTest_Results, DEBUG, INFO, ERROR, WARNING, CRITICAL
from time import time, sleep
from sbc_gpio.gpio_libs._generic_gpio import GpioOut

TEST_NAME = 'LED Flash GPIOD'
TEST_DESCRIPTION = 'Flash an LED connected to a GPIO using the GPIOD library'
PASS_THRESHOLD = 1

class DevTest_LED(DevTest_Base):
    ''' Class for an LED flash using gpiod 
        Requires a GpioPin class for cross platform support.
        Calling class should pass platform appropriate derivitive '''
    def __init__(self, gpio:GpioOut, log_level=INFO):
        super().__init__(log_level)
        self.__pin = gpio

    def close(self):
        self.__pin.set_low()
        self.__pin.close()
    
    @property
    def is_test_ok(self):
        return True

    def _start_thread_iterations(self, iterations, on_ms=1000, off_ms=1000, run_secs=None):
        ''' Run the blink test for a specified number of iterations '''
        with self._testing_lock:
            iter_run, iter_success = 0, 0
            start_time = time()
            while iter_run < iterations and not self._stop_tests:
                try:
                    self.__pin.set_high()
                    sleep(on_ms / 1000)
                    self.__pin.set_low()
                    sleep(off_ms / 1000)
                    iter_success += 1
                except Exception as e:
                    self._logger.error(f"{self.info_str}: Error flashing LED: {e}")
                finally:
                    iter_run += 1
            end_time = time()
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success, 
                                                parameters={'iterations': iterations, 'on_ms': on_ms, 'off_ms': off_ms,
                                                            'gpio': str(self.__pin)},
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION)

    def _start_thread_time(self, run_secs=10, iterations=None, on_ms=1000, off_ms=1000):
        ''' Run the blink test for a specified number of seconds '''
        with self._testing_lock:
            iter_run, iter_success = 0, 0
            start_time = time()
            stop_time = time() + run_secs
            while time() < stop_time and not self._stop_tests:
                try:
                    self.__pin.set_high()
                    sleep(on_ms / 1000)
                    self.__pin.set_low()
                    sleep(off_ms / 1000)
                    iter_success += 1
                except Exception as e:
                    self._logger.error(f"{self.info_str}: Error flashing LED: {e}")
                finally:
                    iter_run += 1
            end_time = time()
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success, 
                                                parameters={'run_secs': run_secs, 'on_ms': on_ms, 'off_ms': off_ms,
                                                            'gpio': str(self.__pin)},
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION)


'''
GPIO library:  RPi.GPIO
Supported platforms: Raspberry Pi devices only
'''
import RPi.GPIO as GPIO
from sbc_gpio import PULL, EVENT
from ._generic_gpio import GpioIn as Generic_GpioIn, GpioOut as Generic_GpioOut
from logging_handler import INFO
from threading import Thread
from time import time, sleep
from datetime import timedelta


# initialization
GPIO.setmode(GPIO.BCM) # type: ignore
GPIO.setwarnings(False) # type: ignore


class GpioOut(Generic_GpioOut):
    ''' Class to represent an abstracted GPIO pin using the RPi.GPIO library '''
    def __init__(self, gpio_pin, gpio_chip=0, name=None, pull=PULL.NONE, log_level=INFO, initial_state=0):
        super().__init__(name=name, log_level=log_level, pull=pull)
        self.name = name if name is not None else f"pin:{gpio_pin}"
        self.gpio_pin, self.gpio_chip = gpio_pin, gpio_chip
        GPIO.setup(int(gpio_pin), GPIO.OUT) # type: ignore

    def close(self):
        pass

    @property
    def state(self):
        ''' return curent CS state '''
        return GPIO.input(self.gpio_pin) # type: ignore
    
    def set_high(self):
        ''' Set the pin to on/high '''
        GPIO.output(self.gpio_pin, 1) # type: ignore

    def set_low(self):
        ''' Set the pin to off/low '''
        GPIO.output(self.gpio_pin, 0) # type: ignore



class GpioIn(Generic_GpioIn):
    ''' Class to represent an abstracted GPIO pin using the RPi.GPIO '''
    def __init__(self, gpio_pin, gpio_chip=0, name=None, pull=PULL.DOWN, event=EVENT.BOTH, debounce_ms=100, callback=None, log_level=INFO, start_polling=True):
        super().__init__(name=name, log_level=log_level, event=event, callback=callback, debounce_ms=debounce_ms, pull=pull)
        self.name = name if name is not None else f"chip:{gpio_chip},pin:{gpio_pin}"
        self.gpio_pin, self.gpio_chip = gpio_pin, gpio_chip
        # initialize the chip and pin
        GPIO.setup(int(gpio_pin), GPIO.IN, pull_up_down=GPIO.PUD_UP if pull == PULL.UP else (GPIO.PUD_DOWN if pull == PULL.DOWN else None)) # type: ignore

        if start_polling:
            self.start()

    def close(self):
        pass

    @property
    def state(self):
        ''' Return current CS state '''
        return GPIO.input(self.gpio_pin) # type: ignore
    
    def _event_thread(self): # type: ignore
        ''' Background thread to watch for rising or falling edge '''
        # NOTE: The RPi.GPIO library includes a debounce however it doesn't include separate rising and falling
        triggered = False

        def call_event(event, triggered):
            if self.callback is not None:
                self._logger.debug(f"{self.info_str}: {'RISING' if event else 'FALLING'} state: {triggered}")
                Thread(target=self.callback, kwargs={'event': EVENT.RISING if event else EVENT.FALLING,
                        'time': time(), 'state': triggered}).start()
            else:
                self._logger.info(f"{self.info_str}: {'RISING' if event else 'FALLING'} state: triggered")


        while True and not self._stop_thread:
            try:
                event_triggered = GPIO.wait_for_edge(self.gpio_pin, GPIO.BOTH, bouncetime=self.debounce_ms, timeout=1000) # type: ignore
                triggered_state = self.state
                if event_triggered is not None:
                    # if monitoring for event rising or event falling ONLY, just send the event
                    if self.event != EVENT.BOTH:
                        if (self.event == EVENT.RISING and triggered_state) or (self.event == EVENT.FALLING and not triggered_state):
                            triggered = True
                            call_event(triggered_state, triggered)
                    else:
                        if (triggered_state and self.pull == PULL.DOWN) or (not triggered_state and self.pull != PULL.DOWN):
                            triggered = True
                            call_event(triggered_state, triggered)
                        elif triggered:
                            triggered = False
                            call_event(triggered_state, triggered)

            except Exception as e:
                self._logger.error(f"{self.info_str}: Error in rising event thread: {e}. Restarting...")
                triggered = False
                sleep(.5)

        # reset the stop thread variable
        self._stop_thread = False



'''
These classes are used to represent a dynamic overlay.  The dynamic overlay is a workaround
for SBC's that use extlinux for bootng.  Unlike the typical method you may be used to with a
Raspberry Pi, extlinux doesn't support passing parameters to an overlay.  All overlays must be
compiled as complete DTB files.  The Dynamic Overlay will compile a DTB after generating a
DTS file using the parameters from the /boot/config.txt (default for most platforms) file.

Only a limited number of drivers are supported.  Check the platform specific configuration
for the list of supported dynamic overlays.

For comparison:

RPI-4 overlay to enable gpio-ir-tx on a GPIO:
============================================
(in /boot/config.txt)
dtoverlay=gpio-ir-tx,gpio_pin=18

Using dynamic overlay on a extlinux system: (Rock Pi 5B used as an example)
==========================================
(in /boot/config.txt)
dynamic_overlay=gpio-ir-tx,gpio=3b3

After adding the dynamic Overlay to the /boot/config.txt file, run "update_dynamic_overlay.py"
to generate the DTS and DTBO files.  For each platform, before editing the extlinux.conf file
we need to run the platform specific tool to run their update on extlinux.conf first.  For
exeample, "sudo update_extlinux.sh" is used to regenerate the extlinux.conf file from the
/boot/config.txt parameters.  THIS WILL REMOVE ALL DYNAMIC OVERLAYS!  The 
"update_dynamic_overlay.py" tool will run "update_extlinux.sh" first, then add the dynamic
overlays after the extlinux.conf file is updated.

NOTE:  When using dynamic overlays, be sure to use "update_dynamic_overlay.py" INSTEAD of
the platform specific "update_extlinux.sh" command, otherwise your dynamic overlays will be
lost on next boot.

NOTE!!!  "update_dynamic_overlay,py" MUST be run as root or with sudo!  This will read the
contents of the /boot/config.txt file, create DTS/DTBO files in "/boot/dynamic_overlay" and
write changes to "/boot/extlinux/extlinux.conf".

How to run:
==========
sudo update_dynamic_overlay.py
(this is installed in your path as a script with the install. If you installed the package
using a venv, be sure to active the venv first!)
'''

import logging

class OverlayParam:
    ''' Represent a parameter in an overlay '''
    def __init__(self, name:str, param_type:str, number_as_str=False, param_help='No additional help available', values=None, default_value=None, set_value=None):
        self.name = name
        if param_type in ['str', 'string']:
            self.param_type = str
        elif param_type in ['int', 'integer']:
            self.param_type = int
        elif param_type in ['bool', 'boolean']:
            self.param_type = bool
        elif param_type in ['float', 'decimal']:
            self.param_type = float
        else:
            self.param_type = None
        self.number_as_str = number_as_str
        self.param_help = param_help
        self.values = values if values is not None else []
        self.default_value = default_value
        self.set_value = default_value if set_value is None else set_value

    def __str__(self):
        ''' Print the parameter help as a line '''
        return f"({'required' if self.default_value is None else 'optional'}) " +  \
                str(self.values) if self.values else '' + \
                f"(DEFAULT: {self.default_value})" if self.default_value else '' + \
                self.param_help

    def update_value(self, value):
        ''' Update the set value.  Verify type and if in values list '''
        if self.param_type is not None:
            if self.number_as_str and isinstance(value, int) or isinstance(value, float):
                value = str(value)
            if not isinstance(value, self.param_type):
                raise ValueError(f"Unable to set value for {self.name} to {value}. Does not match type {self.param_type}")
        if len(self.values) != 0:
            if value not in self.values:
                raise ValueError(f"Unable to set value for {self.name} to {value}. Supported values: {self.values}")
        self.set_value = value

    @property
    def ok(self):
        ''' Return True if parameter value is set '''
        return self.set_value is not None


class DynamicOverlay:
    ''' Represent an overlay '''
    def __init__(self, name:str, template='gpio_basic.dts.j2', params=None):
        ''' Create overlay instance from the configuration file '''
        self.name = name
        self.template = template
        self.params = []
        if isinstance(params, list):
            for _param in params:
                try:
                    self.params.append(OverlayParam(**_param))
                except Exception as ex:
                    logging.error('Unable to import parameter for %s: %s', self.name, _param)
                    raise ex
        else:
            raise ValueError(f"Expecting a list of parameters and received {type(params)}")

    @property
    def ok(self) -> bool:
        ''' Return True if all required parameters are configured, else False '''
        return len(self.missing_params) == 0

    @property
    def missing_params(self) -> list:
        ''' Return a list of required parameters that are missing '''
        missing_params = []
        for _param in self.params:
            if not _param.ok:
                missing_params.append(_param.name)
        return missing_params

    @property
    def param_name_list(self) -> list:
        ''' Return a list of all available parameters '''
        return [_overlay_param.name for _overlay_param in self.params]

    def get_param(self, param_name:str) -> OverlayParam:
        ''' Return a specific parameter '''
        for _param in self.params:
            if _param.name == param_name:
                return _param
        raise ValueError(f'Unknown parameter {param_name}.  Parameters are {self.param_name_list}')

    def set_params(self, params:dict):
        ''' Update the params for the overlay.  Verify all params set '''
        for item, value in params.items():
            if item not in self.param_name_list:
                logging.error('Parameter %s not a valid parameter', item)
                raise ValueError(f"Paramter {item} not a valid parameter")
            self.get_param(item).update_value(value)
        for _param in self.params:
            if not _param.ok:
                raise ValueError(f"Not all required parameters configured. {_param} is required")



from ._test_base import DevTest_Base, DevTest_Results, DEBUG, INFO, ERROR, WARNING, CRITICAL
import os
from serial import Serial
import string
from time import time, sleep
import random

TEST_NAME = 'UART'
TEST_DESCRIPTION = 'Uses CP2102 connected to USB and UART RX/TX pins to test serial communication'
PASS_THRESHOLD = 0.75

SER_BAUDRATES = [9600, 115200, 230400, 460800, 576000, 921600]
SER_BLOCK_SIZES = [64, 128, 256, 512, 1024, 2048]
SER_SEND_DELAY_ADD = .20

class DevTest_UART(DevTest_Base):
    ''' Class for a serial test using UART and a CP2102 '''
    def __init__(self, uart_interface, usb_interface, log_level=INFO, baud_rates=SER_BAUDRATES, block_sizes=SER_BLOCK_SIZES, 
                 send_delay=SER_SEND_DELAY_ADD):
        super().__init__(log_level)
        self._uart_interface = uart_interface
        self._usb_interface = usb_interface
        self._uart_ok = None
        self._uart_dev = None
        self._usb_dev = None
        self.baud_rates = baud_rates
        self.block_sizes = block_sizes
        self.send_delay = send_delay

    def close(self):
        if isinstance(self._uart_dev, Serial):
            self._logger.info(f"{self.info_str}: Closing {self._uart_interface}...")
            self._uart_dev.close()
        if isinstance(self._usb_dev, Serial):
            self._logger.info(f"{self.info_str}: Closing {self._usb_interface}...")
            self._usb_dev.close()

    @property
    def is_test_ok(self):
        if self._uart_ok is not None:
            return self._uart_ok
        try:
            dev_folder = os.listdir('/dev')            
            # check if uart and usb interfaces are present
            if self._uart_interface in dev_folder:
                self._uart_dev = Serial(port='/dev/' + self._uart_interface)
            else:
                raise NameError(f"Device {self._uart_interface} not found!")
                
            if self._usb_interface in dev_folder:
                self._usb_dev = Serial(port='/dev/' + self._usb_interface)
            else:
                raise NameError(f"Device {self._usb_interface} not found!")

            if not isinstance(self._uart_dev, Serial) or not isinstance(self._usb_dev, Serial):
                raise TypeError(f"Unable to initiate Serial interfaces!")
        except Exception as e:
            self._logger.error(f"{self.info_str}: {e}")
            self._uart_dev = None
            self._usb_dev = None
            return False
        return True

    

    def _start_thread_time(self, run_secs, iterations=None):
        if isinstance(self._uart_dev, Serial) and isinstance(self._usb_dev, Serial):
            with self._testing_lock:
                # clear receive queues
                self._uart_dev.read_all()
                self._usb_dev.read_all()
                random_char_list = string.ascii_letters + string.digits
                sent_rates = [[0] * len(SER_BAUDRATES), [0] * len(SER_BAUDRATES)]   # [sent using ser1, send using ser2]
                sent_sizes = [[0] * len(SER_BLOCK_SIZES), [0] * len(SER_BLOCK_SIZES)]   # [sent using ser1, send using ser2]
                recv_rates = [[0] * len(SER_BAUDRATES), [0] * len(SER_BAUDRATES)]  # [received on ser2, received on ser1]
                recv_sizes = [[0] * len(SER_BLOCK_SIZES), [0] * len(SER_BLOCK_SIZES)]  # [received on ser2, received on ser1]
                iter_run = 0
                iter_success = 0
                start_time = time()
                stop_time = time() + run_secs
                while time() < stop_time and not self._stop_tests:
                    for baud_rate in self.baud_rates:
                        # set the serial baud rates
                        self._uart_dev.baudrate = baud_rate
                        self._usb_dev.baudrate = baud_rate
                        
                        for block_size in self.block_sizes:
                            # generate random data and send on both serial interfaces
                            random_data = ''.join([random.choice(random_char_list) for x in range(block_size)]).encode() + b'\n'
                            self._uart_dev.write(random_data)
                            sent_rates[0][self.baud_rates.index(baud_rate)] += 1
                            sent_sizes[0][self.block_sizes.index(block_size)] += 1
                            self._usb_dev.write(random_data)
                            sent_rates[1][self.baud_rates.index(baud_rate)] += 1
                            sent_sizes[1][self.block_sizes.index(block_size)] += 1
                            iter_run += 2
                            sleep((block_size * 8 / baud_rate) + self.send_delay)

                            # read the serial data
                            usb_recv_data = self._usb_dev.read_all()
                            uart_recv_data = self._uart_dev.read_all()

                            # record success / failures
                            if usb_recv_data == random_data:
                                recv_rates[0][self.baud_rates.index(baud_rate)] += 1
                                recv_sizes[0][self.block_sizes.index(block_size)] += 1
                                self._logger.debug(f"{self.info_str}: Sent from {self._uart_interface} ({baud_rate}baud/{block_size}chars) OK")
                                iter_success += 1
                            else:
                                self._logger.warning(f"{self.info_str}: Sent from {self._uart_interface} ({baud_rate}baud/{block_size}chars) FAILED")
                            if uart_recv_data == random_data:
                                recv_rates[1][self.baud_rates.index(baud_rate)] += 1
                                recv_sizes[1][self.block_sizes.index(block_size)] += 1
                                self._logger.debug(f"{self.info_str}: Sent from {self._usb_interface} ({baud_rate}baud/{block_size}chars) OK")
                                iter_success += 1
                            else:
                                self._logger.warning(f"{self.info_str}: Sent from {self._usb_interface} ({baud_rate}baud/{block_size}chars) FAILED")

                            if time() >= stop_time or self._stop_tests:
                                break
                        if time() >= stop_time or self._stop_tests:
                            break
                end_time = time()

                test_details = []
                test_details.append(f"{self._uart_interface}: ({sum(recv_rates[0])}/{sum(sent_rates[0])}) (baud:recv/sent): {', '.join([f'{self.baud_rates[i]}:{recv_rates[0][i]}/{sent_rates[0][i]}' for i in range(len(self.baud_rates))])}")
                test_details.append(f"{self._usb_interface}: ({sum(recv_rates[1])}/{sum(sent_rates[1])}) (baud:recv/sent): {', '.join([f'{self.baud_rates[i]}:{recv_rates[1][i]}/{sent_rates[1][i]}' for i in range(len(self.baud_rates))])}")
                test_details.append(f"{self._uart_interface}: ({sum(recv_sizes[0])}/{sum(sent_sizes[0])}) (bs:recv/sent): {', '.join([f'{self.block_sizes[i]}:{recv_sizes[0][i]}/{sent_sizes[0][i]}' for i in range(len(self.block_sizes))])}")
                test_details.append(f"{self._usb_interface}: ({sum(recv_sizes[1])}/{sum(sent_sizes[1])}) (bs:recv/sent): {', '.join([f'{self.block_sizes[i]}:{recv_sizes[1][i]}/{sent_sizes[1][i]}' for i in range(len(self.block_sizes))])}")
                self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success,
                                                    parameters={'run_secs': run_secs, 'baud_rates': self.baud_rates, 'block_sizes': self.block_sizes},
                                                    start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD, name=TEST_NAME,
                                                    description=TEST_DESCRIPTION, details=test_details)


    _start_thread_iterations = _start_thread_time # type: ignore
    



import unittest
from logging_handler import create_logger, INFO, DEBUG
from time import time, sleep

from sbc_gpio import SBCPlatform, EVENT

from sbc_gpio.device_tests.ir import DevTest_IR
from sbc_gpio.device_tests.led_gpiod import DevTest_LED
from sbc_gpio.device_tests.button_gpiod import DevTest_Button
from sbc_gpio.device_tests.dht_spi import DevTest_DHT
from sbc_gpio.device_tests.bmx_spi import DevTest_BMX
from sbc_gpio.device_tests.i2c_display import DevTest_I2CDisp
from sbc_gpio.device_tests.uart import DevTest_UART

logger = create_logger(INFO, name='tester')

gpio_tests = {
    'Rock5B': {
        'led': '3A7',
        'btn': '3B6',
        'dht': '1D7',
        'dht_spi': 0,
        "dht22": True,
        'bmx': '3B5',
        'bmx_spi': 1,
        'i2c': 7,
        "ir": True,
        "uart_dev": "ttyS2",
        "usb_dev": "ttyUSB1"
    },
    'Pi4B': {
        "led": 17,
        "btn": 27,
        "dht": 5,
        "dht22": True,
        "dht_spi": 0,
        "bmx": 16,
        "bmx_spi": 1,
        "i2c": 1,
        "ir": True,
        "uart_dev": "ttyS0",
        "usb_dev": "ttyUSB0"
    }
}

platform = SBCPlatform()
test_vars = dict(gpio_tests.get(platform.model, {}))


class devtesterTest(unittest.TestCase):
    def test_1_ir_test(self):
        logger.info('===================================== %s', self._testMethodName)
        dev_tester = DevTest_IR(log_level=INFO)
        self.assertTrue(dev_tester.is_test_ok)
        logger.info('----------------- Test 10 iterations blocking')
        dev_tester.start(iterations=10, wait=True)
        logger.info(f'{dev_tester.test_results}')
        self.assertTrue(dev_tester.test_results.passed)
        logger.info('----------------- Test 10 seconds blocking')
        dev_tester.start(run_secs=30, wait=True)
        logger.info(f'{dev_tester.test_results}')
        self.assertTrue(dev_tester.test_results.passed)

    def test_2_led_test(self):
        logger.info('===================================== %s', self._testMethodName)
        pin = platform.get_gpio_out(test_vars.get('led'))
        led_tester = DevTest_LED(pin, log_level=INFO)
        logger.info('----------------- Test 10 iterations blocking')
        led_tester.start(iterations=10, wait=True, on_ms=250, off_ms=750)
        logger.info(f'{led_tester.test_results}')
        self.assertTrue(led_tester.test_results.passed)
        logger.info('----------------- Test 10 seconds blocking')
        led_tester.start(run_secs=10, wait=True, on_ms=750, off_ms=250)
        logger.info(f'{led_tester.test_results}')
        self.assertTrue(led_tester.test_results.passed)

    def test_3_button_test(self):
        logger.info('===================================== %s', self._testMethodName)
        pin = platform.get_gpio_in(test_vars.get('btn'), event=EVENT.RISING)
        button_tester = DevTest_Button(pin, log_level=INFO)
        logger.info('----------------- Test 10 iterations blocking')
        button_tester.start(iterations=10, run_secs=10, wait=True)
        logger.info(f'{button_tester.test_results}')
        self.assertTrue(button_tester.test_results.passed)
        logger.info('----------------- Test 10 seconds blocking')
        button_tester.start(run_secs=10, wait=True)
        logger.info(f'{button_tester.test_results}')
        self.assertTrue(button_tester.test_results.passed)
        pin.close()

    def test_4_dht(self):
        logger.info('===================================== %s', self._testMethodName)
        pin = platform.convert_gpio_tuple(test_vars.get('dht'))
        dht_tester = DevTest_DHT(int(test_vars.get('dht_spi', 0)), pin, log_level=INFO, dht22=test_vars.get('dht22', False))
        logger.info('----------------- Test 10 iterations blocking')
        dht_tester.start(iterations=10, wait=True)
        logger.info(f'{dht_tester.test_results}')
        self.assertTrue(dht_tester.test_results.passed)
        logger.info('----------------- Test 10 seconds blocking')
        dht_tester.start(run_secs=10, wait=True)
        logger.info(f'{dht_tester.test_results}')
        self.assertTrue(dht_tester.test_results.passed)

    def test_5_bmx(self):
        logger.info('===================================== %s', self._testMethodName)
        pin = platform.convert_gpio_tuple(test_vars.get('bmx'))
        bmx_tester = DevTest_BMX(int(test_vars.get('bmx_spi', 1)), pin, log_level=INFO)
        logger.info('----------------- Test 10 iterations blocking')
        bmx_tester.start(iterations=10, wait=True)
        logger.info(f'{bmx_tester.test_results}')
        self.assertTrue(bmx_tester.test_results.passed)
        logger.info('----------------- Test 10 seconds blocking')
        bmx_tester.start(run_secs=10, wait=True)
        logger.info(f'{bmx_tester.test_results}')
        self.assertTrue(bmx_tester.test_results.passed)

    def test_6_i2c(self):
        logger.info('===================================== %s', self._testMethodName)
        i2c_tester = DevTest_I2CDisp(test_vars.get('i2c', 1), log_level=INFO)
        logger.info('----------------- Test 10 iterations blocking')
        i2c_tester.start(iterations=10, wait=True)
        logger.info(f'{i2c_tester.test_results}')
        self.assertTrue(i2c_tester.test_results.passed)
        logger.info('----------------- Test 10 seconds blocking')
        i2c_tester.start(run_secs=10, wait=True)
        logger.info(f'{i2c_tester.test_results}')
        self.assertTrue(i2c_tester.test_results.passed)

    def test_7_uart(self):
        logger.info('===================================== %s', self._testMethodName)
        uart_tester = DevTest_UART(test_vars.get('uart_dev'), test_vars.get('usb_dev'), log_level=DEBUG)
        logger.info('----------------- Test 10 iterations blocking')
        uart_tester.start(iterations=10, wait=True)
        logger.info(f'{uart_tester.test_results}')
        for detail in uart_tester.test_results.details:
            logger.info(f"    {detail}")
        self.assertTrue(uart_tester.test_results.passed)
        logger.info('----------------- Test 10 seconds blocking')
        uart_tester.start(run_secs=120, wait=True)
        logger.info(f'{uart_tester.test_results}')
        for detail in uart_tester.test_results.details:
            logger.info(f"    {detail}")
        self.assertTrue(uart_tester.test_results.passed)

    def test_7_threading(self):
        logger.info('===================================== %s', self._testMethodName)
        bmx_pin = platform.convert_gpio_tuple(test_vars.get('bmx'))
        bmx_tester = DevTest_BMX(1, bmx_pin, log_level=INFO)
        dht_pin = platform.convert_gpio_tuple(test_vars.get('dht'))
        dht_tester = DevTest_DHT(0, dht_pin, log_level=INFO, dht22=True)
        btn_pin = platform.get_gpio_in(test_vars.get('btn'), event=EVENT.RISING)
        btn_tester = DevTest_Button(btn_pin, log_level=INFO)
        led_pin = platform.get_gpio_out(test_vars.get('led'))
        led_tester = DevTest_LED(led_pin, log_level=INFO)
        ir_tester = DevTest_IR(log_level=INFO)
        uart_tester = DevTest_UART(test_vars.get('uart_dev'), test_vars.get('usb_dev'), log_level=INFO)
        i2c_tester = DevTest_I2CDisp(test_vars.get('i2c'), log_level=INFO)
        # verify all tests are ok to run
        self.assertTrue(bmx_tester.is_test_ok and dht_tester.is_test_ok and btn_tester.is_test_ok and led_tester.is_test_ok and ir_tester.is_test_ok and i2c_tester.is_test_ok)
        # start all the threads for 30 seconds
        run_secs=30
        bmx_tester.start(run_secs=run_secs)
        dht_tester.start(run_secs=run_secs)
        btn_tester.start(run_secs=run_secs)
        led_tester.start(run_secs=run_secs)
        ir_tester.start(run_secs=run_secs)
        i2c_tester.start(run_secs=run_secs)
        uart_tester.start(run_secs=run_secs)
        stop_time = time() + run_secs
        while time() < stop_time:
            logger.info(f'Tasks Running:  BMX: {bmx_tester.is_running}, DHT: {dht_tester.is_running}, BTN: {btn_tester.is_running}, LED: {led_tester.is_running}, IR: {ir_tester.is_running}, I2C: {i2c_tester.is_running}')
            sleep(1)
        while bmx_tester.is_running or dht_tester.is_running or btn_tester.is_running or led_tester.is_running or ir_tester.is_running or i2c_tester.is_running:
            logger.info(f'Tasks Running:  BMX: {bmx_tester.is_running}, DHT: {dht_tester.is_running}, BTN: {btn_tester.is_running}, LED: {led_tester.is_running}, IR: {ir_tester.is_running}, I2C: {i2c_tester.is_running}')
            sleep(1)
        self.assertFalse(bmx_tester.is_running and dht_tester.is_running and btn_tester.is_running and led_tester.is_running and ir_tester.is_running and i2c_tester.is_running)
        logger.info(f"BMX results:  {bmx_tester.test_results}")
        logger.info(f"DHT results:  {dht_tester.test_results}")
        logger.info(f"BTN results:  {btn_tester.test_results}")
        logger.info(f"LED results:  {led_tester.test_results}")
        logger.info(f"IR  results:  {ir_tester.test_results}")
        logger.info(f"I2C results:  {i2c_tester.test_results}")
        logger.info(f"UART Results: {uart_tester.test_results}")
        for detail in uart_tester.test_results.details:
            logger.info(f"    {detail}")
        self.assertTrue(bmx_tester.test_results.passed and dht_tester.test_results.passed and btn_tester.test_results.passed and led_tester.test_results.passed and ir_tester.test_results.passed and i2c_tester.test_results.passed and uart_tester.test_results.passed)



'''
Base test classes used to represent the test results as well as a structure for the tests.

'''
import logging
from logging_handler import create_logger, DEBUG, INFO, ERROR, WARNING, CRITICAL
from time import time, sleep
import threading


class DevTest_Results:
    ''' 
    Base class to represent the results of a test.

    Properties:
      iterations: (int) total number of tests that were run
      iter_success: (int) number of successful tests that were run
      parameters: (dict) parameters that were used for the test
      start_time: (int) unix time when test was started
      end_time: (int) unix time when test completed
      pass_threshold: (int) percentage required to qualify as successful
      pass_on_zero: (bool) allow a pass if zero successful iterations
      name: (str) Name of the test
      description: (str) Descrption of the test
      details: [list of str] Additional testing data
      passed: (bool) true/false if met pass threshold
      pass_percent: (float) decimal rounded to 2 for the pass rate

    Methods:
    str(DevTest_Results) - Prints the results of the test as a string
    bool(DevTest_Results) - returns True/False if the test passed
    '''
    iterations = None
    iter_success = None
    parameters = {}
    start_time = None
    end_time = None
    pass_threshold = 0
    pass_on_zero = False
    name = ''
    description = ''
    details= []

    def __init__(self, **kwargs):
        ''' 
        All passed kwargs are assigned to the local class fields. The following are used:
        iterations (number of tests run), iter_success (number of successful runs), parameters
        (dict of test parameters), start_time (unix start time), end_time (unix end time),
        pass_threshold (percentage required for test to pass), pass_on_zero (allow a pass if
        there were zero successful iterations), name (name of test), description (description 
        of test), details (list of strings, additional test detailed output if needed) 
        '''
        for key, value in kwargs.items():
            self.__setattr__(key, value)

    @property
    def passed(self):
        ''' return True/False if the test passed '''
        if self.end_time and self.iterations and self.iter_success and self.pass_threshold:
            return self.iter_success / self.iterations >= self.pass_threshold
        if self.pass_on_zero and self.pass_threshold == 0:
            return True
        return False
    
    @property
    def pass_percent(self):
        ''' Return the pass percentage '''
        if self.end_time and self.iterations and self.iter_success and self.pass_threshold:
            return round(self.iter_success / self.iterations, 2)
        if self.pass_on_zero and self.pass_threshold == 0:
            return 1
        return 0
        
    def __str__(self):
        return f"Test {self.name} {'PASSED' if self.passed else 'FAILED'}: {self.iter_success} / {self.iterations} iterations successful. " \
               f"{self.pass_percent * 100}%, pass theshold is {self.pass_threshold * 100}%"
    
    def __bool__(self):
        return self.passed


class DevTest_Base:
    ''' Base class to represent a test.  Will be overriden be specific test cases '''
    def __init__(self, log_level=INFO):
        self._logger = create_logger(console_level=log_level, name=self.info_str)
        self._stop_tests = False
        self._testing_lock = threading.Lock()
        self._testing_thread = None

    def __del__(self):
        self.close()

    def close(self):
        ''' Close and clean up any objects '''
        pass

    @property
    def info_str(self):
        """ Returns the info string for the class (used in logging commands) """
        return f"{self.__class__.__name__}"

    @property
    def is_running(self):
        ''' Return True/False if a test is running '''
        return self._testing_lock.locked()
    
    def stop(self, wait=False, timeout=10):
        ''' Stop any test in process '''
        if self.is_running:
            self._logger.info(f"{self.info_str}: Sent stop request to test. Timeout {timeout}...")
            self._stop_tests = True
            if wait:
                cancel_time = time() + timeout
                while self.is_running:
                    sleep(.25)

    @property
    def is_test_ok(self):
        ''' Check if the system is able to run the test '''
        pass

    def config_test(self):
        ''' Provide configuration to the test (may also be passed during init)'''
        pass


    def start(self, run_secs=10, iterations=None, wait=False, asyncio=False, **kwargs):
        ''' Start the test '''
        if not self.is_test_ok:
            raise ImportError(f"{self.info_str}: Unable to start test")
        self.test_results = None
        if iterations is not None:
            self._logger.info(f"{self.info_str}: Running test with {iterations} iterations. Wait: {wait}. Other Args: {kwargs}")
            if wait:
                self._start_thread_iterations(iterations=iterations, run_secs=run_secs, **kwargs)
            else:
                if not self.is_running:
                    pass_kwargs = {'iterations': iterations, 'run_secs': run_secs}
                    pass_kwargs.update(kwargs)
                    self._testing_thread = threading.Thread(target=self._start_thread_iterations, kwargs=pass_kwargs, name=self.info_str)
                    self._testing_thread.start()
                else:
                    self._logger.warning(f"{self.info_str}: Unable to start threaded test as another test is currently running.")
        else:
            self._logger.info(f"{self.info_str}: Running test for {run_secs} seconds. Wait: {wait}. Other Args: {kwargs}")
            if wait:
                self._start_thread_time(run_secs=run_secs, **kwargs)
            else:
                if not self.is_running:
                    pass_kwargs = {'iterations': iterations, 'run_secs': run_secs}
                    pass_kwargs.update(kwargs)
                    self._testing_thread = threading.Thread(target=self._start_thread_time, kwargs=pass_kwargs, name=self.info_str)
                    self._testing_thread.start()
                else:
                    self._logger.warning(f"{self.info_str}: Unable to start threaded test as another test is currently running.")

    def _start_thread_iterations(self, iterations, run_secs):
        pass

    def _start_thread_time(self, run_secs, iterations):
        pass


import sbc_gpio
from logging_handler import create_logger, INFO, DEBUG

logger = create_logger(DEBUG, name='tester')

platform = sbc_gpio.SBCPlatform(DEBUG)
print(platform.model)

platform.write_dynamic_overlays()
print('test123')


'''
PyPi Package: sbc_gpio
Homepage: https://www.learningtopi.com/python-modules-applications/sbc_gpio/
Git: https://github.com/LearningToPi/sbc_gpio

Description:
===========
Calling the module directly can be used for either printing the model and serial number
of the SBC device as it is identified, or executing a series of tests based on test classes
available in the sbc_gpio/device_tests folder.

Executing tests requires a config file that includes the GPIO info needed for each of the tests.
A sample configuration file can be generated using the CLI options outlined below:


Usage Example:
=============

>> Output the platform info
$ python3 -m sbc_gpio
2023-04-30 22:59:55,337 - SBCPlatform - INFO - SBCPlatform: Platform identified as Rock5B (Radxa ROCK 5B)
###################################
SBC Model:        Rock5B
SBC Description:  Radxa ROCK 5B
SBC Serial:       ***************
SPI Buses:        1, 0
I2C Buses:        7, 6, 4, 1, 0, 9
###################################

>> Create a configuration file
$ python3 -m sbc_gpio --write-config --config configs/test.json
Sample configuration written to 'configs/test.json'.

>> Execute a test using a configuration file
$ python3 -m sbc_gpio --config configs/rock5b.json  --time 30
2023-04-30 23:05:19,705 - SBCPlatform - INFO - SBCPlatform: Platform identified as Rock5B (Radxa ROCK 5B)
2023-04-30 23:05:19,706 - GpioOut - INFO - GpioOut (chip:3,pin:7): Requesting GPIO...
2023-04-30 23:05:19,707 - GpioIn - INFO - GpioIn (chip:3,pin:14): Requesting GPIO...
2023-04-30 23:05:19,707 - GpioIn - INFO - GpioIn (chip:3,pin:14): Starting event thread...
2023-04-30 23:05:19,708 - dht11_spi - INFO - DHT22_Spi (SPI0-500.0kHz): Opening
2023-04-30 23:05:19,708 - dht11_spi - INFO - DHT22_Spi (SPI0-500.0kHz): Setting SPI bus speed from 50000000 to 500000
2023-04-30 23:05:19,709 - root - INFO - Bmx280Spi (SPI1.13): Opening
<...>
2023-04-30 23:05:53,880 - SBC_Tester - INFO - Test LED Flash GPIOD PASSED: 15 / 15 iterations successful. 100.0%, pass theshold is 100%
2023-04-30 23:05:53,882 - SBC_Tester - INFO - Test Button GPIOD PASSED: 0 / 0 iterations successful. 100%, pass theshold is 0%
2023-04-30 23:05:53,882 - SBC_Tester - INFO - Test DHT11 over SPI PASSED: 13 / 14 iterations successful. 93.0%, pass theshold is 75.0%
2023-04-30 23:05:53,882 - SBC_Tester - INFO - Test BMP280/BME280 over SPI PASSED: 28 / 28 iterations successful. 100.0%, pass theshold is 75.0%
2023-04-30 23:05:53,883 - SBC_Tester - INFO - Test I2C Display PASSED: 27 / 27 iterations successful. 100.0%, pass theshold is 100%
2023-04-30 23:05:53,883 - SBC_Tester - INFO - Test IR GPIO TX/RX PASSED: 22 / 22 iterations successful. 100.0%, pass theshold is 75.0%
2023-04-30 23:05:53,883 - SBC_Tester - INFO - Test UART PASSED: 108 / 108 iterations successful. 100.0%, pass theshold is 75.0%
2023-04-30 23:05:53,883 - SBC_Tester - INFO -     ttyS2: (54/54) (baud:recv/sent): 9600:12/12, 115200:12/12, 230400:12/12, 460800:6/6, 576000:6/6, 921600:6/6
2023-04-30 23:05:53,884 - SBC_Tester - INFO -     ttyUSB1: (54/54) (baud:recv/sent): 9600:12/12, 115200:12/12, 230400:12/12, 460800:6/6, 576000:6/6, 921600:6/6
2023-04-30 23:05:53,884 - SBC_Tester - INFO -     ttyS2: (54/54) (bs:recv/sent): 64:9/9, 128:9/9, 256:9/9, 512:9/9, 1024:9/9, 2048:9/9
2023-04-30 23:05:53,884 - SBC_Tester - INFO -     ttyUSB1: (54/54) (bs:recv/sent): 64:9/9, 128:9/9, 256:9/9, 512:9/9, 1024:9/9, 2048:9/9
2023-04-30 23:05:53,884 - SBC_Tester - INFO - Completed test run for 30 seconds.
2023-04-30 23:05:53,963 - DevTest_IR - INFO - DevTest_IR: lircd TX process pid: 282170 stopping...
2023-04-30 23:05:54,071 - DevTest_IR - INFO - DevTest_IR: lircd RX process pid: 282187 stopping...
2023-04-30 23:05:55,243 - GpioOut - INFO - GpioOut (chip:3,pin:7): Releasing GPIO...
2023-04-30 23:05:55,244 - GpioOut - INFO - GpioOut (chip:3,pin:7): Releasing GPIO...
2023-04-30 23:05:55,244 - DevTest_UART - INFO - DevTest_UART: Closing ttyS2...
2023-04-30 23:05:55,244 - DevTest_UART - INFO - DevTest_UART: Closing ttyUSB1...

    
MIT License

Copyright (c) 2022 LearningToPi <contact@learningtopi.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''
import argparse
import os
import json
import sys
from time import sleep
from logging_handler import create_logger, INFO, WARNING
from sbc_gpio.device_tests.led_gpiod import DevTest_LED
from sbc_gpio.device_tests.button_gpiod import DevTest_Button
from sbc_gpio.device_tests.i2c_display import DevTest_I2CDisp
from sbc_gpio.device_tests.ir import DevTest_IR
from sbc_gpio.device_tests.dht_spi import DevTest_DHT
from sbc_gpio.device_tests.bmx_spi import DevTest_BMX
from sbc_gpio.device_tests.uart import DevTest_UART
from . import SBCPlatform


def run_test(run_secs=60, led=None, btn=None, dht=None, ir=None, dht_spi=None, dht22=False, bmx=None, bmx_spi=None, i2c=None, log_level=INFO,
             uart_dev=None, usb_dev=None):
    ''' Run a basic set of tests on the specified devices.  All tests are run in parallel for a number of seconds.'''
    logger = create_logger(console_level=log_level, name='SBC_Tester')

    # loop through and start each test
    platform = SBCPlatform()
    # only run a test if passed values are not None or empty string
    tests = []
    if led is not None and led != '':
        if platform.gpio_is_valid(led):
            tests.append(DevTest_LED(gpio=platform.get_gpio_out(led), log_level=log_level))
    if btn is not None and btn != '':
        if platform.gpio_is_valid(btn):
            tests.append(DevTest_Button(gpio=platform.get_gpio_in(btn), log_level=log_level))
    if dht is not None and dht != '' and isinstance(dht_spi, int):
        if platform.gpio_is_valid(dht):
            tests.append(DevTest_DHT(spi_bus=dht_spi, dht22=dht22, gpio_tuple=platform.convert_gpio_tuple(dht), log_level=log_level))
    if bmx is not None and bmx != '' and isinstance(bmx_spi, int):
        if platform.gpio_is_valid(bmx):
            tests.append(DevTest_BMX(spi_bus=bmx_spi, gpio_tuple=platform.convert_gpio_tuple(bmx), log_level=log_level))
    if isinstance(i2c, int) and i2c in platform.i2c_buses():
        tests.append(DevTest_I2CDisp(port=i2c, log_level=log_level))
    if isinstance(ir, bool) and ir:
        tests.append(DevTest_IR(log_level=log_level))
    if uart_dev is not None and usb_dev is not None:
        tests.append(DevTest_UART(uart_dev, usb_dev, log_level=log_level))

    # start the tests
    for test in tests:
        logger.info(f'Starting test {test.info_str}')
        test.start(run_secs=run_secs)

    try:
        logger.info('Waiting for tests to complete...')
        sleep(run_secs)

        # check to make sure all tests complete before continuing
        for x in range(10):
            for test in tests:
                if test.is_running:
                    sleep(1)
                    continue

        # print out if any tests failed to complete and results for the completed tests
        for test in tests:
            if test.is_running:
                logger.error(f"Test {test.infostr} failed to complete!  Results not available.")
            else:
                logger.info(test.test_results)
                if test.test_results.details is not None:
                    for line in test.test_results.details:
                        logger.info(f"    {line}")

        logger.info(f'Completed test run for {run_secs} seconds.')
    except KeyboardInterrupt:
        # if Ctrl+C stop all tests
        logger.warning('Keyboard Interrupt caught. Stopping all tests. This may take a few seconds to complete...')
        for test in tests:
            test.stop()
        for x in range(10):
            for test in tests:
                if test.is_running:
                    sleep(1)
                    continue


if __name__ == '__main__':
    # setup the argument parser
    parser = argparse.ArgumentParser(description="Execute a sequence of tests on the SBC GPIO's or if no arguments print the SBC system data.")
    parser.add_argument('--time', required=False, type=int, default=60, help="(60) Number of seconds to run the test")
    parser.add_argument('--config', required=False, type=str, help="Config file to read from or write to")
    parser.add_argument('--output', required=False, type=str, help="Output file for results of the test run")
    parser.add_argument('--write-config', required=False, action='store_true', default=False, help="(False) Write a sample config file (requires config parameter)")
    parser.add_argument('--log-level', dest='log_level', required=False, type=str, default='INFO', help='(INFO) Specify the logging level for the console (DEBUG, INFO, WARN, CRITICAL)')

    args = vars(parser.parse_args())

    if args.get('write_config', False) and args.get('config') is None:
        sys.tracebacklimit=0
        raise ValueError('Parameter "write-config" requires parameter "config"')

    if args.get('write_config', False):
        sample_config = {
            'led': '3A7',
            'btn': '3B6',
            'dht': '1D7',
            'dht22': True,
            'dht_spi': 0,
            'bmx': '3B5',
            'bmx_spi': 1,
            'i2c': 7,
            'ir': True,
            "uart_dev": "ttyS0",
            "usb_dev": "ttyUSB0"
        }
        if os.path.isfile(args.get('config', 'sample-config.json')):
            print(f"Config file '{args.get('config', 'sample-config.json')}' already exists")
            overwrite = input("Overwrite file? (N)o/(Y)es: ")
            if overwrite != "Y":
                print("Quitting without writing configuration file.")
                quit(1)
        with open(args.get('config', 'sample-config.json'), 'w', encoding='utf-8') as output_file:
            output_file.write(json.dumps(sample_config, indent=4, default=str))
            print(f"Sample configuration written to '{args.get('config', 'sample-config.json')}'.")
            quit(0)

    if args.get('config', None) is not None:
        with open(args.get('config', 'sample-config.json'), 'r', encoding='utf-8') as input_file:
            config = json.loads(input_file.read())
        run_test(run_secs=args.get('time', 60), **config)
        quit()

    # print the SBC data to the screen
    platform = SBCPlatform(log_level=WARNING)
    print('#' * 35)
    print('SBC Model:       ', platform.model)
    print('SBC Description: ', platform.description)
    print('SBC Serial:      ', platform.serial)
    print('SPI Buses:       ', ', '.join([str(x) for x in platform.spi_buses()]))
    print('I2C Buses:       ', ', '.join([str(x) for x in platform.i2c_buses()]))
    print('#' * 35)



'''
Platform specific file for the Radxa Rock 5B.  This system uses extlinux for boot and supports
dynamic overlays.  GPIO's are provided in the format <chip><bank><number> (i.e. "0A3", "3B3").
'''
import os
import re
from collections import namedtuple
import jinja2
from sbc_gpio import PLATFORM_INFO
from sbc_gpio.dynamic_dts import DynamicOverlay
from ._generic import set_gpio_flags

# select the gpio library for the platform
from sbc_gpio.gpio_libs.gpiod import GpioIn, GpioOut #pylint: disable=W0611,C0411

MODEL_IDENTIFIER = [
    {
        'type': 'file', 'file': '/sys/firmware/devicetree/base/model', 'contents': 'Radxa ROCK 5B'
    }
]

DYNAMIC_OVERLAYS={
    "gpio-ir-tx": {
        "params": [
            {
                "name": "gpio",
                "param_type": "str",
                "param_help": "String to identify the gpio. i.e. 14, 112, 2A6",
                "number_as_str": True
            },
            {
                "name": "active",
                "param_type": "str",
                "values": ["high", "low"],
                "param_help": "Set active [high|low]",
                "default_value": "high"

            },
            {
                "name": "pull",
                "param_type": "str",
                "values": ["high", "low", "none"],
                "param_help": "Set internal pull-up/pull-down or no bias [high|low|none]",
                "default_value": "none"
            }
        ],
        "template": "gpio-basic.dts"
    },
    "gpio-ir-recv": {}
}

SERIAL_NUMBER = '/sys/firmware/devicetree/base/serial-number'
PLATFORM_LOCAL = namedtuple('PLATFORM_LOCAL', ('gpio_re_format', 'serial_path', 'gpio_prefix'))
PLATFORM_SPECIFIC = PLATFORM_INFO(model='Rock5B',
              description='Radxa ROCK 5B',
              gpio_valid_values=(139,138,115,113,111,112,42,41,43,150,63,47,103,110,13,14,109,100,148,44,45,149,114,105,106,107),
              dynamic_overlay_dir='/boot/dynamic_overlay',
              dynamic_overlays=DYNAMIC_OVERLAYS,
              config_file='/boot/config.txt',
              update_extlinux_script='/usr/local/sbin/update_extlinux.sh',
              extlinux_conf='/boot/extlinux/extlinux.conf',
              local=PLATFORM_LOCAL(gpio_re_format="^(?P<chip>[0-4]?)(?P<pinprefix>[ABCD])(?P<pinnum>[0-7])$",
                                   serial_path='/sys/firmware/devicetree/base/serial-number',
                                   gpio_prefix=('A', 'B', 'C', 'D')))



def convert_gpio(gpio_str:str) -> int:
    ''' convert a gpio passed as a string to an integer '''
    gpio_tuple = convert_gpio_tuple(gpio_str=gpio_str)
    gpio_int = gpio_tuple[0] * 32 + gpio_tuple[1]
    if gpio_int not in PLATFORM_SPECIFIC.gpio_valid_values:
        raise ValueError(f'Gpio {gpio_int} not in valid range {PLATFORM_SPECIFIC.gpio_valid_values} for device {PLATFORM_SPECIFIC.model}')
    return gpio_tuple[0] * 32 + gpio_tuple[1]

def convert_gpio_tuple(gpio_str:str) -> tuple:
    ''' Take a string representing a GPIO and return it as a Tuple -> ([int chip], [int num])'''
    match = re.search(PLATFORM_SPECIFIC.local.gpio_re_format, gpio_str.upper())
    if not match:
        raise ValueError('GPIO Format for Rock 5B is 0A0-3D8.  GPIO chip (0-4) followed by A-D and 0-8 to identify the pin on the GPIO chip. See https://wiki.radxa.com/Rock5/hardware/5b/gpio')
    if '' in match.groups():
        raise ValueError('GPIO Format for Rock 5B is 0A0-3D8.  GPIO chip (0-4) followed by A-D and 0-8 to identify the pin on the GPIO chip. See https://wiki.radxa.com/Rock5/hardware/5b/gpio')
    return int(match.group('chip')), PLATFORM_SPECIFIC.local.gpio_prefix.index(match.group('pinprefix')) * 8 + int(match.group('pinnum'))


def write_dynamic_overlay(overlay:DynamicOverlay):
    ''' Take an dynamic overlay and write it to the dynamic overlay folder '''
    # write the DTS file from the template
    file_loader = jinja2.FileSystemLoader(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates'))
    env = jinja2.Environment(loader=file_loader)
    template = env.get_template(overlay.template)

    # Create dict of flags that need to be set
    gpio_flags = []
    for param_name in overlay.param_name_list:
        _param = overlay.get_param(param_name)
        if _param.name == 'active':
            if _param.set_value == 'high':
                gpio_flags.append('ACTIVE_HIGH')
            elif _param.set_value == 'low':
                gpio_flags.append('ACTIVE_LOW')
        elif _param.name == 'pull':
            if _param.set_value == 'none':
                gpio_flags.append('PULL_NONE')
            elif _param.set_value == 'high':
                gpio_flags.append('PULL_HIGH')
            elif _param.set_value == 'low':
                gpio_flags.append('PULL_LOW')

    with open(os.path.join(PLATFORM_SPECIFIC.dynamic_overlay_dir, f"{overlay.name}-{overlay.get_param('gpio').set_value}.dts"), 'w', encoding='utf-8') as output_file:
        output_file.write(template.render(driver=overlay.name,
                                          gpio_number=convert_gpio(str(overlay.get_param('gpio').set_value)),
                                          gpio_chip=convert_gpio_tuple(str(overlay.get_param('gpio').set_value))[0],
                                          gpio_pin=convert_gpio_tuple(str(overlay.get_param('gpio').set_value))[1],
                                          gpio_pin_ctrl=set_gpio_flags(*gpio_flags)))



from ._test_base import DevTest_Base, DevTest_Results, DEBUG, INFO, ERROR, WARNING, CRITICAL
from time import time, sleep
from bmx280_spi import Bmx280Spi

TEST_NAME = 'BMP280/BME280 over SPI'
TEST_DESCRIPTION = 'Read from a DHT11 or DHT22 sensor using the SPI bus with dht11_spi library'
PASS_THRESHOLD = .75

class DevTest_BMX(DevTest_Base):
    ''' Class for a DHT11 / DHT22 sensor using the dht11_spi library.
        The SPI bus is used to read the signals from the DHT sensor (vastly improved reliability over bit banging gpio library) '''
    def __init__(self, spi_bus:int, gpio_tuple:tuple, log_level=INFO):
        super().__init__(log_level)
        self._gpio_tuple = gpio_tuple
        self._bmx = Bmx280Spi(spiBus=spi_bus, cs_chip=self._gpio_tuple[0], cs_pin=self._gpio_tuple[1])

    def close(self):
        pass
    
    @property
    def is_test_ok(self):
        return True

    def _start_thread_iterations(self, iterations, run_secs=None, interval=1):
        ''' Run the blink test for a specified number of iterations '''
        with self._testing_lock:
            iter_run, iter_success = 0, 0
            start_time = time()
            while iter_run < iterations and not self._stop_tests:
                try:
                    reading = self._bmx.update_readings()
                    if reading is not None:
                        iter_success += 1
                        self._logger.info(f"{self.info_str}: {self._bmx.model} reading: {reading}")
                    else:
                        self._logger.warning(f"{self.info_str}: Error reading BMX")
                except Exception as e:
                    self._logger.warning(f"{self.info_str}: Error reading BMX: {e}")
                finally:
                    iter_run += 1
                    sleep(interval)
            end_time = time()
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success, 
                                                parameters={'iterations': iterations, 'interval': interval,
                                                            'gpio': str(self._gpio_tuple)},
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION)

    def _start_thread_time(self, run_secs=10, iterations=None, interval=1):
        ''' Run the blink test for a specified number of seconds '''
        with self._testing_lock:
            iter_run, iter_success = 0, 0
            start_time = time()
            stop_time = time() + run_secs
            while time() < stop_time and not self._stop_tests:
                try:
                    reading = self._bmx.update_readings()
                    if reading is not None:
                        iter_success += 1
                        self._logger.info(f"{self.info_str}: {self._bmx.model} reading: {reading}")
                    else:
                        self._logger.warning(f"{self.info_str}: Error reading BMX")
                except Exception as e:
                    self._logger.error(f"{self.info_str}: Error reading BMX: {e}")
                finally:
                    iter_run += 1
                    sleep(interval)
            end_time = time()
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success, 
                                                parameters={'run_secs': run_secs, 'interval': interval,
                                                            'gpio': str(self._gpio_tuple)},
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION)



import unittest
from logging_handler import create_logger, INFO, DEBUG
from time import sleep

import sbc_gpio

logger = create_logger(INFO, name='tester')

gpio_tests = {
    'Rock5B': {
        'gpio_string': ['4B3', '3C2', '3A4', '0B5', '1B1'],
        'gpio_out': '3A7',
        'gpio_in': '3B6'
    }
}


class devtesterTest(unittest.TestCase):
    def test_1_get_platform(self):
        logger.info('===================================== %s', self._testMethodName)
        platform = sbc_gpio.SBCPlatform(INFO)
        logger.info(f"Device Platform: {platform}")
        self.assertTrue(isinstance(platform, sbc_gpio.SBCPlatform))
        logger.info(f"Valid GPIO's: {platform.gpio_valid_values}")
        platform_tests = gpio_tests[platform.model]
        for gpio_string in platform_tests['gpio_string']:
            gpio_val = platform.convert_gpio(gpio_string)
            logger.info(f"GPIO {gpio_string} converts to {gpio_val}. GPIO Valid? {gpio_val in platform.gpio_valid_values}")
            self.assertTrue(gpio_val in platform.gpio_valid_values)
        # test gpio output and input
        gpio_in_events = 0
        def in_callback(**kwargs):
            nonlocal gpio_in_events
            logger.info(f"Event: {kwargs.get('event')} State: {kwargs.get('state')}")
            gpio_in_events += 1
        gpio_in = platform.get_gpio_in(platform_tests['gpio_in'], event='rising', callback=in_callback)
        gpio_out = platform.get_gpio_out(platform_tests['gpio_out'])
        logger.info('Press the input button >= 5 times...')
        logger.info('LED Output will run for 15 seconds...')
        for i in range(15):
            gpio_out.set_high()
            sleep(.5)
            gpio_out.set_low()
            sleep(.5)
        self.assertGreaterEqual(gpio_in_events, 5)
        gpio_in.stop()
        del gpio_in
        del gpio_out



'''
Platform specific file for the Raspberry Pi 4B.  This system does NOT use extlinux for boot so
dynamic overlays are not needed or supported.  GPIO's are provided as a simple GPIO # (i.e. 14, 21).
This configuration should also apply to other Raspberry Pi family devices and may be updated in
the future to include other platforms.
'''

from collections import namedtuple
from sbc_gpio import PLATFORM_INFO

# select the gpio library for the platform
from sbc_gpio.gpio_libs.rpi_gpio import GpioIn, GpioOut #pylint: disable=W0611

MODEL_IDENTIFIER = [
    {
        'type': 'file', 'file': '/sys/firmware/devicetree/base/model', 'contents': 'Raspberry Pi 4 Model B'
    }
]

SERIAL_NUMBER = '/sys/firmware/devicetree/base/serial-number'

PLATFORM_LOCAL = namedtuple('PLATFORM_LOCAL', ('serial_path'))
PLATFORM_SPECIFIC = PLATFORM_INFO(model='Pi4B',
              description='Raspberry Pi 4 Model B',
              gpio_valid_values=[0, 1, 2, 3, 4, 5, 6, 7 , 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27],
              dynamic_overlay_dir='/boot/dynamic_overlay',
              dynamic_overlays=None,
              config_file='/boot/config.txt',
              update_extlinux_script=None,
              extlinux_conf=None,
              local=PLATFORM_LOCAL(serial_path='/sys/firmware/devicetree/base/serial-number'))



'''
Any scripts created during the install of the sbc_gpio package will execute from here.

How to run update_dynamic_overlay():
====================================
sudo update_dynamic_overlay.py
(this is installed in your path as a script with the install)
NOTE: If you installed this module in a venv, sudo will fail.  You will need to install the
module under root to make sure the script is available to run.
'''

from . import SBCPlatform



def update_dynamic_overlay():
    ''' Initiate a call to start the dynamic overlay generation '''
    device = SBCPlatform()
    device.write_dynamic_overlays()



from ._test_base import DevTest_Base, DevTest_Results, DEBUG, INFO, ERROR, WARNING, CRITICAL
from time import time, sleep
from sbc_gpio.gpio_libs._generic_gpio import GpioIn
from threading import Lock

TEST_NAME = 'Button GPIOD'
TEST_DESCRIPTION = 'Test an button connected to a GPIO'
PASS_THRESHOLD = 0

class DevTest_Button(DevTest_Base):
    ''' Class for a button using gpiod with edge monitoring.
        Requires a GpioPin class for cross platform support. 
        Calling class should pass platform appropriate derivitive '''
    def __init__(self, gpio:GpioIn, log_level=INFO):
        super().__init__(log_level)
        self._pin = gpio
        self._pin.callback = self.__button_callback
        self._events = []
        self._event_lock = Lock()

    def close(self):
        self._pin.close()
        
    @property
    def is_test_ok(self):
        return True

    def __button_callback(self, event, **kwargs):
        with self._event_lock:
            self._logger.info(f"Input event {event}, state: {kwargs.get('state')}")
            self.__events.append((kwargs.get('time'), event))

    def _start_thread_iterations(self, run_secs=10, iterations=None, interval=.5):
        ''' Run the thread checking for button events '''
        with self._testing_lock:
            with self._event_lock:
                # clear any events
                self.__events = []
            stop_time = time() + run_secs
            start_time = time()
            while time() < stop_time and not self._stop_tests:
                with self._event_lock:
                    if iterations is not None and len(self.__events) >= iterations:
                        break
                sleep(interval)
            end_time = time()
            with self._event_lock:
                iter_run = len(self.__events)
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_run,
                                                parameters={'run_secs': run_secs, 'iterations': iterations, 'interval': interval},
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION, pass_on_zero=True)
        
    def _start_thread_time(self, *args, **kwargs):
        self._start_thread_iterations(*args, **kwargs)



from ._test_base import DevTest_Base, DevTest_Results, DEBUG, INFO, ERROR, WARNING, CRITICAL
from time import time, sleep
from RPLCD.i2c import CharLCD

TEST_NAME = 'I2C Display'
TEST_DESCRIPTION = 'Connect to an I2C display and print data to the screen'
PASS_THRESHOLD = 1

class DevTest_I2CDisp(DevTest_Base):
    ''' Class for an I2C display. 
        Required Parameters:
            port (int) - I2C bus where the display is located
         '''
    def __init__(self, port, address=0x27, cols=16, rows=2, charmap='A00', auto_linebreaks=True, backlight_enabled=True, i2c_expander='PCF8574', log_level=INFO):
        super().__init__(log_level)
        self._params = {
            'port': port,
            'address': address,
            'cols': cols,
            'rows': rows,
            'charmap': charmap,
            'auto_linebreaks': auto_linebreaks,
            'backlight_enabled': backlight_enabled,
            'i2c_expander': i2c_expander
        }
        self._i2c_display = CharLCD(**self._params)
        
    def close(self):
        self._i2c_display.close()
    
    @property
    def is_test_ok(self):
        return True

    def _start_thread_iterations(self, iterations, interval=1, run_secs=None):
        ''' Run the i2c display test for a specified number of iterations '''
        with self._testing_lock:
            iter_run, iter_success = 0, 0
            start_time = time()
            self._i2c_display.clear()
            self._i2c_display.write_string('LearningToPi.com\n\rIteration: ')
            while iter_run < iterations and not self._stop_tests:
                try:
                    self._i2c_display.cursor_pos = (0, 0)
                    self._i2c_display.write_string(f'LearningToPi.com\n\rIteration: {iter_run + 1}')
                    iter_success += 1
                except Exception as e:
                    self._logger.error(f"{self.info_str}: Error Writing to I2C Display: {e}")
                finally:
                    iter_run += 1
                    sleep(interval)
            end_time = time()
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success, 
                                                parameters={'iterations': iterations, 'interval': interval}.update(self._params),
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION)

    def _start_thread_time(self, run_secs=10, iterations=None, interval=1):
        ''' Run the i2c display test for a specified number of seconds '''
        with self._testing_lock:
            iter_run, iter_success = 0, 0
            start_time = time()
            stop_time = start_time + run_secs
            self._i2c_display.clear()
            self._i2c_display.write_string('LearningToPi.com\n\rIteration: ')
            while time() < stop_time and not self._stop_tests:
                try:
                    self._i2c_display.cursor_pos = (0, 0)
                    self._i2c_display.write_string(f'LearningToPi.com\n\rIteration: {iter_run + 1}')
                    iter_success += 1
                except Exception as e:
                    self._logger.error(f"{self.info_str}: Error Writing to I2C Display: {e}")
                finally:
                    iter_run += 1
                    sleep(interval)
            end_time = time()
            self.test_results = DevTest_Results(iterations=iter_run, iter_success=iter_success, 
                                                parameters={'run_secs': run_secs, 'interval': interval}.update(self._params),
                                                start_time=start_time, end_time=end_time, pass_threshold=PASS_THRESHOLD,
                                                name=TEST_NAME, description=TEST_DESCRIPTION)